class ZZZError(Exception):
    def __init__(self, error_message: dict):
        super().__init__(error_message["text"])
        self.code = error_message["code"]
        self.text = error_message["text"]


import asyncio
import json
import os
import difflib
from typing import List, Optional,Tuple,Union
from .http import AioSession
from .error import ZZZError
from ..model.base import Lang, HoyoAPIHeaders, ErrorText
from ..model.api import CharacterDataHoYo, ZenkaApi

API_URL = "https://enka.network/api/zzz/uid/{uid}"
RAW_BASE_URL = "https://raw.githubusercontent.com/EnkaNetwork/API-docs/master/store/zzz/"
RAW_BASE_URL_ZENKA = "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/json/"
HOYO_MENU = "https://sg-wiki-api-static.hoyolab.com/hoyowiki/zzz/wapi/home/hot_content?menu_id=8"
HOYO_WIKI = "https://sg-wiki-api-static.hoyolab.com/hoyowiki/zzz/wapi/entry_page?entry_page_id={pid}"

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
JSON_DIR = os.path.join(BASE_DIR, "..", "jsons")

FORCE_LOAD_FILES = {
    "level_format.json",
    "star_format.json", 
    "discs_format.json"
}
JSON_FILES = [
    "avatars.json",
    "equipments.json",
    "locs.json",
    "medals.json",
    "namecards.json",
    "pfps.json",
    "property.json",
    "titles.json",
    "weapons.json",
    "hoyolink.json",
    "level_format.json",
    "star_format.json",
    "discs_format.json",
]

jsons_data = {}
lang: str = Lang.EN.value

async def fetch_json(filename: str):
    if filename in ["level_format.json", "star_format.json", "hoyolink.json", "discs_format.json"]:
        if filename == "hoyolink.json":
            return
        url = RAW_BASE_URL_ZENKA + filename
    else:
        url = RAW_BASE_URL + filename
    
    data = await AioSession.get(url, response_format= "json")
    key = filename.replace(".json", "_data")
    jsons_data[key] = data

async def get_json_files():
    tasks = [fetch_json(filename) for filename in JSON_FILES]
    await asyncio.gather(*tasks)
    await fetch_hoyowiki_data()


async def fetch_hoyowiki_data() -> None:
    json_res = await AioSession.get(HOYO_MENU, headers=HoyoAPIHeaders.HEADERS, response_format= "json")
    tasks = [fetch_page(i["ep_abstract"]["entry_page_id"], i["ep_abstract"]["name"]) 
                for i in json_res["data"]["contents"]]
    results = await asyncio.gather(*tasks)
    
    results = [x for x in results if x is not None]

    enka_names = {jsons_data["locs_data"][lang][c_data["Name"]]: cid 
                    for cid, c_data in jsons_data["avatars_data"].items()}
    
    hoyo_data = {name: CharacterDataHoYo(name=name, link=link, id=hcid, faction=faction, type=tipe, element=element)
                    for name, link, hcid, faction, tipe, element in results}

    matches = {}
    unmatched = []

    for b_name in enka_names.keys():
        exact_match = next((h_name for h_name in hoyo_data.keys() if b_name.lower() in h_name.lower()), None)
        
        if exact_match:
            matches[b_name] = exact_match
        else:
            unmatched.append(b_name)

    remaining_hoyo_keys = set(hoyo_data.keys()) - set(matches.values())
    for b_name in unmatched:
        best_match = find_best_match(b_name, list(remaining_hoyo_keys))
        if best_match:
            matches[b_name] = best_match
            remaining_hoyo_keys.remove(best_match) 
    jsons_data["hoyolink_data"] = {
        enka_names[e_name]: hoyo_data[h_name].model_dump()
        for e_name, h_name in matches.items() if h_name
    }

async def fetch_page(pid: str, name: str) -> Union[Tuple[str, str, str, str, str, str], bool]:
    json_res = await AioSession.get(HOYO_WIKI.format(pid = pid), headers=HoyoAPIHeaders.HEADERS, response_format= "json")
    data = json_res["data"]["page"]["filter_values"]

    faction = data.get("agent_faction", {}).get("value_types", [{}])
    if faction:
        faction = faction[0].get("icon", "default_faction_icon")

    if data.get("agent_stats", {})["value_types"]:
        element = data.get("agent_stats", {}).get("value_types", [{}])[0].get("icon", "unknown")
    if data.get("agent_specialties", {})["value_types"]:
        tipe = data.get("agent_specialties", {}).get("value_types", [{}])[0].get("icon", "")
    else:
        return None


    return name, json_res["data"]["page"]["header_img_url"], pid, faction, tipe, element.strip()

def find_best_match(name: str, candidates: List[str]) -> Optional[str]:
    return next((c for c in candidates if name.lower() in c.lower()),
                (difflib.get_close_matches(name, candidates, n=1, cutoff=0.2) or [None])[0])

class DataManager:
    def __init__(self):
        os.makedirs(JSON_DIR, exist_ok=True)

    async def collect_data(self, assets_save: bool = False) -> dict:
        global jsons_data
        if assets_save and self.check_existing_files():
            jsons_data = self.load_local_data()
            return jsons_data

        await get_json_files()
        self.extra_open()

        if assets_save:
            self.save_all_data()
        return jsons_data

    async def update_data(self):
        await get_json_files()
        self.save_all_data()

    def save_all_data(self):
        for key, data in jsons_data.items():
            filename = os.path.join(JSON_DIR, f"{key}.json")
            with open(filename, "w", encoding="utf-8") as f:
                json.dump(data, f, indent=4, ensure_ascii=False)

    def load_local_data(self) -> dict:
        loaded_data = {}
        for filename in JSON_FILES:
            key = filename.replace(".json", "_data")
            filepath = os.path.join(JSON_DIR, filename.replace(".json", "_data.json"))
            if os.path.exists(filepath):
                with open(filepath, "r", encoding="utf-8") as f:
                    loaded_data[key] = json.load(f)
        return loaded_data
    
    def extra_open(self):
        for filename in FORCE_LOAD_FILES:
            key = filename.replace(".json", "_data")
            filepath = os.path.join(JSON_DIR, filename.replace(".json", "_data.json"))
            if os.path.exists(filepath):
                with open(filepath, "r", encoding="utf-8") as f:
                    jsons_data[key] = json.load(f)

    def get_data(self) -> dict:
        return jsons_data

    def check_existing_files(self) -> bool:
        return all(os.path.exists(os.path.join(JSON_DIR, f.replace(".json", "_data.json"))) for f in JSON_FILES)


async def fetch_user(uid: int, original: bool = False) -> ZenkaApi:
    data = await AioSession.get(API_URL.format(uid = uid), response_format= "json")
    if data.get("message"):
        raise ZZZError(ErrorText().format_api(text = data.get("message")))
    
    if original:
        return data
    
    return ZenkaApi(jsons_data=jsons_data, lang=lang, **data["PlayerInfo"])




from pydantic import BaseModel
from typing import List
from PIL import Image
from ..tools.git import ImageCache

_of = ImageCache()
positionTeams = [
    (39,99),
    (844,99),
    (1655,99),
]

crops = {
    3: 0,
    2: 797,
    1: 1594,
}

class ZZZProfileCard(BaseModel):
    color: tuple
    card: Image.Image
    class Config:
        arbitrary_types_allowed = True

class ZZZCard(BaseModel):
    id: int
    name: str
    color: tuple
    icon: str
    card: Image.Image
    
    class Config:
        arbitrary_types_allowed = True

class ZenkaGenerator(BaseModel):
    player: bool = None
    charter_id: List[int] = []
    charter_name: List[str] = []
    cards: List[ZZZCard] = []

   

class ZenkaGeneratorTeams(BaseModel):
    player: bool = None
    charter_id: List[int] = []
    charter_name: List[str] = []
    cards: List[ZZZCard] = []
    builds: List[Image.Image] = []

    class Config:
        arbitrary_types_allowed = True


    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_image_groups(self, agent_id: List[int]) -> List[List[int]]:
        """Split build_id into groups of 3 (like images)"""
        return [agent_id[i:i+3] for i in range(0, len(agent_id), 3)]

    async def build(self, agent_id: List[int], light: bool = True) -> List[Image.Image]:
        """
        Build one or two images with up to 3 cards each
        (based on agent_id list passed in)
        """

        _of.set_mapping(3)

        groups = self.get_image_groups(agent_id)

        if light:
            full_bg = await _of.full_bg
        else:
            full_bg = await _of.full_bg_b

        
        for i, group in enumerate(groups):
            final_img = full_bg.convert("RGBA").copy()
            for j, id_ in enumerate(group):
                card = next((c for c in self.cards if c.id == id_), None)
                if card:
                    final_img.alpha_composite(card.card, positionTeams[j])

            final_img = final_img.crop((0, 0, final_img.width - crops.get(len(group)), final_img.height))
            self.builds.append(final_img)

        return self.builds



import asyncio
from PIL import Image, ImageDraw
from typing import Tuple
from ..model.base import Translator
from ..model.api import Character, Equipment, PlayerData
from ..tools.git import ImageCache
from ..tools import pill
from ..model.generator import ZZZCard
from ..model.base import URL_ASSETS
from ..model.assets import get_agent_icon, get_color_agent

_of = ImageCache()


async def get_skill_icon(index: int) -> Image.Image:
    return {
        0: await _of.skill_1_s,
        1: await _of.skill_4_s,
        2: await _of.skill_2_s,
        3: await _of.skill_5_s,
        5: await _of.skill_6_s,
        6: await _of.skill_3_s,
    }.get(index, await _of.skill_1_s)


async def get_stars(index: int) -> Image.Image:
    return {
        1: await _of.g_one,
        2: await _of.g_two,
        3: await _of.g_three,
        4: await _of.g_four,
        5: await _of.g_five,
    }.get(index, await _of.g_one)


async def get_rarity(index: int) -> Image.Image:
    return {
        2: await _of.rank_w_b,
        3: await _of.rank_w_a,
        4: await _of.rank_w_s,
    }.get(index, await _of.rank_w_b)

async def get_rank_color(index: int) -> Tuple[int,int,int,int]:
    return {
        2: (7,158,254),
        3: (181,7,254),
        4: (254,181,7)
    }.get(index, (254,181,7))

async def get_core(index: int) -> Image.Image:
    return {
        0: await _of.core_skill_a,
        1: await _of.core_skill_b,
        2: await _of.core_skill_c,
        3: await _of.core_skill_d,
        4: await _of.core_skill_e,
        5: await _of.core_skill_f,
    }.get(index, await _of.core_skill_a)


async def get_cinema(index: int) -> Image.Image:
    return {
        0: await _of.conts_1_lvl,
        1: await _of.conts_2_lvl,
        2: await _of.conts_3_lvl,
        3: await _of.conts_4_lvl,
        4: await _of.conts_5_lvl,
        5: await _of.conts_6_lvl,
    }.get(index, await _of.conts_1_lvl)


class StyleOne:
    def __init__(self, data: Character, player: PlayerData, lang: Translator, art: str = None, color: tuple = None, hide: bool = False, crop: int = 0):
        self.data = data
        self.player = player
        self.lang = lang
        self.art = art
        self.hide = hide
        self.crop = crop
        
        if color:
            self.color = color
        elif art:
            self.color = None
        else:
            colors = get_color_agent(data.id)
            self.color = colors.mindscape

        _of.set_mapping(1)
        
    async def create_background(self):
        full_background = Image.new("RGBA", (1927,802), (0,0,0,0))
        image_background = Image.new("RGBA",(1927,802), (0,0,0,0))
        text_frame = await _of.text_frame
        background = await _of.background
        self.background = background.convert("RGBA").copy()
        mask = await _of.mask
        shadow = await _of.shadow

        frame_info = await _of.frame_info
        frame_info = frame_info.copy()

        dark_frame_info = await _of.dark_frame_info
        cadr = await _of.cadr
        frame = await _of.frame
        charter_bg = await _of.charter_bg


        if self.art:
            self.art = await pill.get_user_image(self.art)
            user_image = await pill.get_center_size((563,802), self.art)
            if not self.color:
                self.color = await pill.get_colors(user_image.convert("RGBA"), 5, common=True, radius=30, quality=800)
                ll = await pill.light_level(self.color[:3])
                if ll < 45:
                    self.color = await pill.get_light_pixel_color(self.color[:3],up = True)
            image_background.alpha_composite(user_image)
            full_background.paste(image_background,(0,0), mask.convert("L"))
            self.background.alpha_composite(full_background)
        else:
            text_frame = await pill.recolor_image(text_frame, self.color[:3])
            self.background.alpha_composite(charter_bg)
            self.background.alpha_composite(text_frame, (-2,8))
            icon = get_agent_icon(self.data.id)
            user_image = await pill.get_download_img(icon.image, size=(942,946))
            user_image_w = await pill.recolor_image(user_image, (255,255,255))
            user_image_b = await pill.recolor_image(user_image, (0,0,0))
            image_background.alpha_composite(user_image_w, (-191,0))
            image_background.alpha_composite(user_image_b, (-203,0))
            image_background.alpha_composite(user_image, (-196, 0))
            full_background.paste(image_background,(0,0), mask.convert("L"))
            self.background.alpha_composite(full_background)
        
        await self.create_litetal()
        self.background.alpha_composite(shadow, (379,-17))
        
        frame_info = await pill.recolor_image(frame_info.copy(), self.color[:3])
        cadr = await pill.recolor_image(cadr.copy(), self.color[:3]) 

        frame_info.alpha_composite(dark_frame_info)
        frame_info.alpha_composite(cadr)


        self.background.alpha_composite(frame_info, (398,0))
        self.background.alpha_composite(self.literal, (410,399))
        self.background.alpha_composite(frame, (389, 0))

        self.white_color = (255,255,255,255) if await pill.is_white_text_readable(self.color, 120) else (32,32,32,32)

    async def create_name(self):
        self.name = Image.new("RGBA", (491,142), (0,0,0,0))

        name_main = await pill.create_image_with_text(self.data.name, 69, stroke_width=2, stroke_fill= (0,0,0,255), max_width= 492)
        name = await pill.create_image_with_text(self.data.name, 69, max_width= 492)
        name = await pill.recolor_image(name, self.color[:3]) 
        lvl = f"{self.lang.lvl}: {self.data.level}/{self.data.max_level*10}"
        lvl_main = await pill.create_image_with_text(lvl, 50, stroke_width=2, stroke_fill= (0,0,0,255) , max_width= 492)
        lvl = await pill.create_image_with_text(lvl, 50, max_width= 492)
        lvl = await pill.recolor_image(lvl, self.color[:3]) 
        self.name.alpha_composite(name, (13,9))
        self.name.alpha_composite(name_main, (11,5))
        self.name.alpha_composite(lvl, (17,87))
        self.name.alpha_composite(lvl_main, (15,83))

    async def create_skill(self):
        self.skill = Image.new("RGBA", (363,165), (0,0,0,0))
        skill_coont = await _of.skill_count
        font = await pill.get_font(25)
        last_true_index = next((i for i in reversed(range(len(self.data.cinema))) if self.data.cinema[i]), 0)

        position_skill = [
            (0,0),
            (115,0),
            (230,0),
            (48,90),
            (163,90),
            (269,90),
        ]

        for i, key in enumerate(self.data.skills):
            skill_icon = Image.new("RGBA", (85,75), (0,0,0,0))
            count = skill_coont.copy()
            icon = await get_skill_icon(key.index)
            lvl = key.level
            if last_true_index >= 3 and i != 4:
                lvl += 2
            if last_true_index >= 5 and i != 4:
                lvl += 2
                
            d = ImageDraw.Draw(count)
            x = int(font.getlength(str(lvl)) / 2)
            d.text((22 - x, -3), str(lvl), font=font, fill=self.color)
            
            skill_icon.alpha_composite(icon.resize((75,75)))
            skill_icon.alpha_composite(count,(40,49))
            
            self.skill.alpha_composite(skill_icon, position_skill[i])

    async def create_weapon(self):
        self.weapon = Image.new("RGBA", (458,176), (0,0,0,0))
        if not self.data.weapon:
            return
        
        bg = await _of.weapon_background
        icon = await pill.get_download_img(URL_ASSETS + self.data.weapon.icon, (167,167))
        icon_color = await pill.recolor_image(icon, self.color[:3]) 
        icon_w = await pill.recolor_image(icon, (255,255,255))
        stars = await get_stars(self.data.weapon.cons)

        self.weapon.alpha_composite(bg)
        self.weapon.alpha_composite(icon_w, (0,9))
        self.weapon.alpha_composite(icon_color, (9,0))
        self.weapon.alpha_composite(icon, (5,3))
        self.weapon.alpha_composite(stars.resize((117,42)), (90,124))
        
        name_main = await pill.create_image_with_text(self.data.weapon.name, 31, stroke_width=2, stroke_fill= (0,0,0,255), max_width= 309)
        name = await pill.create_image_with_text(self.data.weapon.name, 31, max_width= 309) 
        name = await pill.recolor_image(name, self.color[:3]) 
        xyz = 0
        if name_main.size[1] > 60:
            xyz = 30
        self.weapon.alpha_composite(name, (157,32-xyz))
        self.weapon.alpha_composite(name_main, (155,30-xyz))
        
        lvl_bg = await _of.levl
        lvl_bg = lvl_bg.copy()

        font = await pill.get_font(24)
        d = ImageDraw.Draw(lvl_bg)
        text_lvl = f"LVL: {self.data.weapon.level}"
        x = int(font.getlength(text_lvl) / 2)
        d.text((47 - x, 3), text_lvl, font=font, fill=self.color)
        x = int(font.getlength(text_lvl) / 2)
        d.text((43 - x, 0), text_lvl, font=font, fill=(255,255,255,255), stroke_width=2, stroke_fill=(0,0,0,255))
        self.weapon.alpha_composite(lvl_bg, (211,123))

        font = await pill.get_font(22)
        up_bg = await _of.up
        up_bg = up_bg.copy()
        d = ImageDraw.Draw(up_bg)
        text_up = f"R{self.data.weapon.cons}"
        x = int(font.getlength(text_up) / 2)
        d.text((17 - x, 3), text_up, font=font, fill=(37,37,37,255))
        self.weapon.alpha_composite(up_bg, (313,123))


        icon_main = await pill.get_download_img(self.data.weapon.main.icon, size=(30,30))
        icon_sub = await pill.get_download_img(self.data.weapon.sub.icon, size=(30,30))
        font = await pill.get_font(27)
        d = ImageDraw.Draw(self.weapon)

        self.weapon.alpha_composite(icon_main, (181,82))
        self.weapon.alpha_composite(icon_sub, (286,82))
        d.text((218, 80), str(self.data.weapon.main.value), font=font, fill=(255,255,255,255))
        formatted_percentage = self.data.weapon.sub.get_value()
        d.text((321, 80), str(formatted_percentage), font=font, fill=(255,255,255,255))

        rarity = await get_rarity(self.data.weapon.rarity)
        self.weapon.alpha_composite(rarity.convert("RGBA").resize((35,35)), (22,18))

    async def create_disc(self, data: Equipment):
        echo_background = Image.new("RGBA", (193,212), (0,0,0,0))
        
        background = Image.new("RGBA", (193,209), (0,0,0,0))
        icon_disc = await pill.get_download_img(URL_ASSETS + data.icon, (152,152))
        color = await pill.get_colors(icon_disc.convert("RGBA"), 15, common=True, radius=5, quality=800)
        frame_background = await _of.frame_background
        frame_background = await pill.recolor_image(frame_background, color[:3])

        background.alpha_composite(icon_disc, (20,0))
        background.alpha_composite(await _of.frame_background_black)
        background.alpha_composite(frame_background)
        background.alpha_composite(await _of.frame_relict)

        lvl_frame = await _of.lvl
        lvl_frame = lvl_frame.copy()

        font = await pill.get_font(25)
        d = ImageDraw.Draw(lvl_frame)
        x = int(font.getlength(f"+{data.level}") / 2)
        d.text((26-x, -4), f"+{data.level}", font=font, fill=(255,255,255,255))
        background.alpha_composite(lvl_frame, (130,80))

        stats_frame = await _of.stats
        stats_frame.copy()
        icon_main = await pill.get_download_img(data.main[0].icon, (22,22))
        stats_frame.alpha_composite(icon_main, (8,1))

        value = str(data.main[0].get_value())
        font = await pill.get_font(20)
        d = ImageDraw.Draw(stats_frame)
        x = int(font.getlength(value) / 2)
        d.text((76-x, 0), value, font=font, fill=(255,255,255,255))
        background.alpha_composite(stats_frame, (7,80))

        position = [
            (15,126),
            (101,126),
            (15,162),
            (101,162),
        ]

        rolls_line =  Image.new("RGBA", (11,1), color)
        font = await pill.get_font(18)
        for i, key in enumerate(data.sub):
            sub_bg = Image.new("RGBA", (75,26), (0,0,0,0))
            sub_icon = await pill.get_download_img(key.icon, (22,22))
            sub_bg.alpha_composite(sub_icon)
            value = str(key.get_value())
            d = ImageDraw.Draw(sub_bg)
            d.text((24, 0), value, font=font, fill=(255,255,255,255))
            x = 2
            for _ in range(key.rolls-1):
                sub_bg.alpha_composite(rolls_line, (x, 25))
                x += 15
            background.alpha_composite(sub_bg, position[i])

        
        background_w_b = await _of.frame_wb_background
        echo_background.alpha_composite(background_w_b)
        echo_background.alpha_composite(background.resize((187,201)), (3,11))

        rarity = await get_rarity(data.rarity)
        rank_leve = await _of.rank_leve
        rank_leve = await pill.recolor_image(rank_leve, await get_rank_color(data.rarity))
        echo_background.alpha_composite(rank_leve)
        echo_background.alpha_composite(rarity.copy().convert("RGBA").resize((27,27)), (10,16))
        return echo_background

    async def create_litetal(self):
        self.literal = Image.new("RGBA", (222,391))
        
        position = [
            (171,0),
            (137,68),
            (103,136),
            (69,203),
            (36,269),
            (0,340),
        ]
        for i in range(6):
            icon = await get_core(i)
            icon = icon.resize((51,51))
            if self.data.core_skill-1 >= i:
                icon = await pill.recolor_image(icon, self.color[:3])
            else:
                icon = await pill.recolor_image(icon, (32,32,32))
            self.literal.alpha_composite(icon, position[i])

    async def create_cinema(self):
        self.cinema = Image.new("RGBA", (190,346), (0,0,0,0))
        position = [
            (0,0),
            (28,59),
            (55,117),
            (82,176),
            (110,235),
            (137,293),
        ]
        
        for i in range(6):
            icon = await get_cinema(i)
            if i < self.data.const and self.data.const != 0:
                icon = await pill.recolor_image(icon, (0,0,0))
            else:
                icon = await pill.recolor_image(await _of.conts_0_lvl, (0,0,0))
                
            self.cinema.alpha_composite(icon.resize((53,53)),position[i])

    async def create_stats(self):
        self.stats = Image.new("RGBA", (727,296), (0,0,0,0))
        stats_background = await _of.stats_background
        stats_value = await _of.value
        stats_value = await pill.recolor_image(stats_value, self.color[:3])
        xx = 0
        y = 0
        for i, key in enumerate(self.data.stats):
            bg = stats_background.copy()
            bg_val = stats_value.copy()
            icon = await pill.get_download_img(key.icon, (35,35))
            bg.alpha_composite(icon, (10,14))
            name = await pill.create_image_with_text(key.name, 22, 115)
            bg.alpha_composite(name, (49, int(32 - (name.height /2))))
            
            font = await pill.get_font(22)
            d = ImageDraw.Draw(bg_val)
            text_val = key.get_value()
            x = int(font.getlength(text_val) / 2)
            d.text((45 - x, 0), text_val, font=font, fill= self.white_color)
            bg.alpha_composite(bg_val,(0,57))

            d = ImageDraw.Draw(bg)
            font = await pill.get_font(13)

            add = key.get_value(add = True)
            base = key.get_value(base = True)

            d.text((98, 60), str(base), font=font, fill=(255,255,255,255))
            x = int(font.getlength(f"+{add}"))
            d.text((164 - x, 71), f"+{add}", font=font, fill=self.color)

            self.stats.alpha_composite(bg, (xx,y))

            xx += 180
            if i in [3,7]:
                xx = 0
                y += 102

    async def create_sets(self):
        filtered_data = {
            key: {**value, 'count': 2 if value['count'] == 3 else value['count']}
            for key, value in self.sets.items() if value['count'] != 1
        }
        self.sets = []
        sets_background = await _of.sets_background
        sets_collor = await _of.sets_collor
        for key in filtered_data:
            bg = sets_background.copy()
            icon_disc = await pill.get_download_img(URL_ASSETS + filtered_data[key]["icon"], (152,152))
            color = await pill.get_colors(icon_disc.convert("RGBA"), 15, common=True, radius=5, quality=800)
            bg_color = await pill.recolor_image(sets_collor, color[:3])
            bg.alpha_composite(bg_color)
            bg.alpha_composite(await _of.sets_count)
            font = await pill.get_font(27)
            d = ImageDraw.Draw(bg)
            d.text((30, 15),f"{filtered_data[key]['count']}X", font=font, fill=(255,255,255,255))
            font = await pill.get_font(24)
            white_color = (255,255,255,255) if pill.is_white_visible(color) else (32,32,32,32)
            d.text((82, 18),filtered_data[key]['name'], font=font, fill=white_color)

            self.sets.append(bg)

    async def create_nickname(self):
        self.nickname = Image.new("RGBA", (400,72), (0,0,0,0))
        
        name = await pill.create_image_with_text(self.player.profile.nickname, 32, stroke_width=1, stroke_fill= (0,0,0,255), max_width= 250)
        name_c = await pill.recolor_image(name, self.color[:3])

        uid = f"UID: {self.player.profile.uid}"
        if self.hide:
            uid = f"UID: Hide"
        uid = await pill.create_image_with_text(uid, 25,stroke_width=1, stroke_fill= (0,0,0,255), max_width= 250)
        uid_c = await pill.recolor_image(uid, self.color[:3])

        self.nickname.alpha_composite(name_c, (6,6))
        #self.nickname.alpha_composite(name_b, (2,6))
        self.nickname.alpha_composite(name, (4,4))

        self.nickname.alpha_composite(uid_c, (5,40))
        #self.nickname.alpha_composite(uid_b, (2,40))
        self.nickname.alpha_composite(uid, (4,38))

    async def build(self):
        self.background.alpha_composite(self.name, (559,16))
        self.background.alpha_composite(self.skill, (622,184))
        self.background.alpha_composite(self.weapon, (663,395))
        self.background.alpha_composite(self.cinema, (434,25))
        self.background.alpha_composite(self.stats, (1165 - self.crop,10))
        self.background.alpha_composite(self.nickname, (0,724))
        x = 1194 - self.crop
        y = 333
        for _, key in enumerate(self.disc):
            self.background.alpha_composite(key, (x,y))
            x += 243
            if _ == 2:
                x = 1194 - self.crop
                y = 565

        position = {
            1: [(102,71)], 
            2: [(57,18), (146,95)],
            3: [(60,0), (203,69), (0,126)],
        }.get(len(self.sets))

        sets = Image.new("RGBA", (547,195), (0,0,0,0))
        for i, key in enumerate(self.sets):
            sets.alpha_composite(key, position[i])
            pass
        
        self.background.alpha_composite(sets, (541,586))


        if self.crop:
            self.background = self.background.crop((0, 0, self.background.width - self.crop - 20, self.background.height))

    async def start(self) -> ZZZCard:
        await self.create_background()
        await self.create_litetal()
        task = [
            self.create_name(),
            self.create_skill(),
            self.create_weapon(),
            self.create_cinema(),
            self.create_stats(),
            self.create_nickname(),
        ]

        await asyncio.gather(*task)

        self.disc = []
        self.sets = {}
        for key in self.data.equippe:
            self.disc.append(await self.create_disc(key.equipment))
            if not key.equipment.sets.id in self.sets:
                self.sets[key.equipment.sets.id] = {"count": 1, "name": key.equipment.sets.name, "icon": key.equipment.sets.icon}
            else:
                self.sets[key.equipment.sets.id]["count"] += 1

        await self.create_sets()

        await self.build()
        
        return ZZZCard(
            id = self.data.id,
            name = self.data.name,
            color = self.color,
            icon = self.data.icon.circle_icon,
            card = self.background
        )


import random
from PIL import Image, ImageDraw
from ..model.base import Translator
from ..model.api import Character, PlayerData
from ..tools.git import ImageCache
from ..tools import pill
from ..model.generator import ZZZProfileCard
from ..model.base import URL_ASSETS, DEFAULT_AVATAR
from ..model.assets import get_agent_icon, hex_to_rgba

_of = ImageCache()

class StyleOneProfile:
    def __init__(self, data: Character, player: PlayerData, lang: Translator, color: tuple = None, hide: bool = False):
        self.data = data
        self.player = player
        self.lang = lang
        self.hide = hide
        if color:
            self.color = color
        else:
            if player.profile.title:
                self.color = hex_to_rgba(player.profile.title.color.accent)
            else:
                self.color = (235, 174, 89, 255)

        _of.set_mapping(2)

    async def create_medal(self):
        self.medal = []
        font = await pill.get_font(23)
        medal_icon = await _of.medal
        for key in self.player.medal_list:
            medal = medal_icon.copy()
            icon = await pill.get_download_img(URL_ASSETS + key.icon, (84,73))
            medal.alpha_composite(icon)
            d = ImageDraw.Draw(medal)
            x = int(font.getlength(str(key.value)) / 2)
            d.text((107-x, 20), str(key.value), font=font, fill=self.color)

            self.medal.append(medal)

    async def create_background(self):
        self.background = await _of.main_background
        self.background = self.background.copy()
        background = await pill.recolor_image(await _of.background, self.color[:3])

        self.background.alpha_composite(background)
        self.background.alpha_composite(await _of.cinema, (160,296))
        self.background.alpha_composite(await _of.cinema_frame, (160,296))
        self.background.alpha_composite(await _of.texture_avatar, (37,81))

        frame_icon = await _of.frame_icon
        frame_icon = await pill.recolor_image(frame_icon, self.color[:3])
        
        if self.player.profile.icon:
            icon = await pill.get_download_img(URL_ASSETS + self.player.profile.icon, (179,179))
        else:
            icon = await pill.get_download_img(DEFAULT_AVATAR, (179,179))
        self.background.alpha_composite(icon, (58,72))
        self.background.alpha_composite(frame_icon, (56,71))
        self.background.alpha_composite(await _of.frame_avatar, (198,93))

        level = await _of.lvl_b
        lvl_c = await _of.lvl_c
        lvl_c = await pill.recolor_image(lvl_c, self.color[:3])
        level = level.copy()
        level.alpha_composite(await _of.lvl_w)
        level.alpha_composite(lvl_c)
        
        font = await pill.get_font(20)
        d = ImageDraw.Draw(level)
        text_lvl = f"LVL: {self.player.profile.level}"
        x = int(font.getlength(text_lvl) / 2)
        d.text((40 - x, 10), text_lvl, font=font, fill=(33,33,33,255))

        uid_c = await _of.uid_c
        uid_c = await pill.recolor_image(uid_c, self.color[:3])
        uid = await _of.uid_b
        uid_c.alpha_composite(await _of.uid_w)
        uid_c.alpha_composite(uid)

        d = ImageDraw.Draw(uid_c)
        text_uid = f"Agent: {self.player.profile.uid}"
        if self.hide:
            text_uid = f"Agent: Hide"
        x = int(font.getlength(text_uid) / 2)
        d.text((118 - x, 10), text_uid, font=font, fill=self.color)

        self.background.alpha_composite(level, (42,64))
        self.background.alpha_composite(uid_c, (215,32))

        medal_x = 592
        for key in self.medal:
            self.background.alpha_composite(key, (medal_x,95))
            medal_x += 154


        font = await pill.get_font(25)
        d = ImageDraw.Draw(self.background)
        text_desc = self.player.desc
        d.text((274, 200), text_desc, font=font, fill=self.color)

        font = await pill.get_font(35)
        d.text((263, 89), self.player.profile.nickname, font=font, fill=(0,0,0,255))
        d.text((265, 87), self.player.profile.nickname, font=font, fill=(255,255,255,255))

        line = Image.new("RGBA", (1,99), (0,0,0,255))
        self.background.alpha_composite(line, (569,85))
        if self.player.profile.title:
            title = await pill.create_image_with_text(self.player.profile.title.name, 25, 292, (30,30,30,255))
            self.background.alpha_composite(title, (265, int(162 - (title.height/2))))

    async def create_character(self):
        active_position = [
            (-5,134), (40,306), (85,476),
            (208,-36), (253,134), (298,306),
            (474,-36), (519,134), (564,306),
            (744,-36), (786,134), (834,306),
        ]

        mask = await _of.mask
        font = await pill.get_font(23)
        cards = []
        cards_no_lvl = []

        for key in self.player.characters:
            charter_icon = Image.new("RGBA", (143,179), (0,0,0,0))
            charter_icon_mask = Image.new("RGBA", (143,179), (0,0,0,0))

            icon = get_agent_icon(key.id)
            icon = await pill.get_download_img(icon.image, (524,524))

            charter_icon_mask.alpha_composite(icon, (-177,-53))
            charter_icon_mask.alpha_composite(await _of.shadow_charter)
            charter_icon.paste(charter_icon_mask, (0,0), mask.convert("L"))
            cards_no_lvl.append(charter_icon.copy())


            d = ImageDraw.Draw(charter_icon)
            lvl = f"LVL: {key.level}"
            d.text((36, 130), lvl, font=font, fill=(0,0,0,255))
            d.text((38, 130), lvl, font=font, fill=(255,255,255,255))

            cards.append(charter_icon)

        self.charter_icon_total = Image.new("RGBA", (914,622), (0,0,0,0))

        use_position = [
            (-51,-36), (-5,134), (40,306), (85,476),
            (208,-36), (253,134), (298,306), (343,476),
            (474,-36), (519,134), (564,306), (609,476), 
            (744,-36), (786,134), (834,306), (879,476),
        ]
        for key in cards:
            position = random.choice(active_position)
            self.charter_icon_total.alpha_composite(key, position)
            active_position.remove(position)
            use_position.remove(position)

        for position in use_position:
            card = random.choice(cards_no_lvl)
            card = card.convert("LA").convert("RGBA")
            card = await pill.apply_opacity(card, 0.2)
            self.charter_icon_total.alpha_composite(card, position)

    async def start(self) -> ZZZProfileCard:
        await self.create_medal()
        await self.create_background()
        await self.create_character()
        self.background.alpha_composite(self.charter_icon_total, (160,296))

        return ZZZProfileCard(
            color = self.color,
            card = self.background
        )


# Copyright 2025 DEViantUa <t.me/deviant_ua>
# All rights reserved.

import aiohttp
import asyncio
import json
import inspect
import threading
import weakref

class SharedObject:
    __slots__ = (
        'func', 'args', 'kwargs',
        'object',
        'users',
        'sync_lock', 'async_locks',
    )
    
    def __init__(self, func, /, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        
        self.sync_lock = threading.Lock()
        self.async_locks = weakref.WeakKeyDictionary()
        
        self.users = 0
    
    @property
    def async_lock(self, /):
        token = asyncio.get_running_loop()
        
        try:
            lock = self.async_locks[token]
        except KeyError:
            self.async_locks[token] = lock = asyncio.Lock()
        
        return lock
    
    def __enter__(self, /):
        object = None
        
        with self.sync_lock:
            if self.users == 0:
                object = self.func(*self.args, **self.kwargs)
                
                if hasattr(object, '__enter__'):
                    object = object.__enter__()
                
                self.object = object
            else:
                object = self.object
            
            self.users += 1
        
        return object
    
    def __exit__(self, /, *exc_info):
        result = None
        
        with self.sync_lock:
            self.users -= 1
            
            if self.users == 0:
                object = self.object
                
                if hasattr(object, '__exit__'):
                    result = object.__exit__(*exc_info)
                elif hasattr(object, 'close'):
                    object.close()
                
                del self.object
        
        return result
    
    async def __aenter__(self, /):
        object = None
        
        with self.sync_lock:
            async with self.async_lock:
                if self.users == 0:
                    object = self.func(*self.args, **self.kwargs)
                    
                    if inspect.isawaitable(object):
                        object = await object
                    
                    if hasattr(object, '__aenter__'):
                        object = await object.__aenter__()
                    elif hasattr(object, '__enter__'):
                        object = object.__enter__()
                    
                    self.object = object
                else:
                    object = self.object
                
                self.users += 1
        
        return object
    
    async def __aexit__(self, /, *exc_info):
        result = None
        
        with self.sync_lock:
            async with self.async_lock:
                self.users -= 1
                
                if self.users == 0:
                    object = self.object
                    
                    if hasattr(object, '__aexit__'):
                        result = await object.__aexit__(*exc_info)
                    elif hasattr(object, 'aclose'):
                        await object.aclose()
                    elif hasattr(object, 'close'):
                        if inspect.isawaitable(coro := object.close()):
                            await coro
                    elif hasattr(object, '__exit__'):
                        result = object.__exit__(*exc_info)
                    
                    del self.object
        
        return result



class AioSession:
    session = SharedObject(aiohttp.ClientSession)
    proxy = None
    
    @classmethod
    async def enter(cls, proxy = None, /):
        cls.proxy = proxy
        
        return await cls.session.__aenter__()
    
    @classmethod
    async def exit(cls, /, *exc_info):
        if not exc_info:
            exc_info = (None, None, None)
        
        return await cls.session.__aexit__(*exc_info)
    
    @classmethod
    async def creat_session(cls):
        """Creates a session
        Returns:
            aiohttp.ClientSession: The session instance.
        """
        cls.session = aiohttp.ClientSession()
        
        return cls.session
        
    @classmethod
    async def get_session(cls):
        """
        Returns the session instance, creating it if necessary.

        Returns:
            aiohttp.ClientSession: The session instance.
        """
        if cls.session is None:
            cls.session = aiohttp.ClientSession()
        return cls.session

    @classmethod
    async def close_session(cls):
        """
        Closes the session if it exists.
        """
        if cls.session is not None:
            await cls.session.close()
            cls.session = None

    '''@classmethod
    def session(cls):
        """
        Returns the current session instance.

        Returns:
            aiohttp.ClientSession: The current session instance, or None if the session is not created.
        """
        return cls.session'''

    @classmethod
    async def get(cls, url, headers=None, response_format='json', proxy=None, **kwargs):
        """
        Sends a GET request using the current session instance.

        Args:
            url (str): The URL to send the request to.
            headers (dict): Optional headers to include in the request.
            response_format (str): The format of the response data to return.
            **kwargs: Additional arguments to pass to the aiohttp GET method.

        Returns:
            Depends on the specified response format:
            - dict: for 'json' format;
            - str: for 'text' format;
            - bytes: for 'bytes' format.
        """
        async with cls.session as session:
            if not cls.proxy is None:
                proxy = cls.proxy
                
            try:
                async with session.get(url, headers=headers, proxy=proxy, **kwargs) as response:
                    data = await cls.process_response(response, response_format)
                    return data
            except Exception as e:
                print(url)
                print(f"Error during GET request: {e}")


    @classmethod
    async def post(cls, url, data=None, headers=None, response_format='json', proxy=None, **kwargs):
        """
        Sends a POST request using the current session instance.

        Args:
            url (str): The URL to send the request to.
            data: The data to send in the request body.
            headers (dict): Optional headers to include in the request.
            response_format (str): The format of the response data to return.
            **kwargs: Additional arguments to pass to the aiohttp POST method.

        Returns:
            Depends on the specified response format:
            - dict: for 'json' format;
            - str: for 'text' format;
            - bytes: for 'bytes' format.
        """
        if not cls.proxy is None:
            proxy = cls.proxy
                
        async with cls.session as session:
            try:
                async with session.post(url, data=data, headers=headers, proxy=proxy, **kwargs) as response:
                    data = await cls.process_response(response, response_format)
                    return data
            except Exception as e:
                print(f"Error during POST request: {e}")

    @classmethod
    async def request(cls, method, url, headers=None, response_format='json', proxy=None, **kwargs):
        """
        Sends a custom HTTP request using the current session instance.

        Args:
            method (str): The HTTP method to use (e.g., 'GET', 'POST').
            url (str): The URL to send the request to.
            headers (dict): Optional headers to include in the request.
            response_format (str): The format of the response data to return.
            **kwargs: Additional keyword arguments to pass to the request function.

        Returns:
            Depends on the specified response format:
            - dict: for 'json' format;
            - str: for 'text' format;
            - bytes: for 'bytes' format.
        """
        if not cls.proxy is None:
            proxy = cls.proxy
            
        async with cls.session as session:
            try:
                async with session.request(method, url, headers=headers, proxy=proxy, **kwargs) as response:
                    data = await cls.process_response(response, response_format)
                    return data
            except Exception as e:
                print(f"Error during custom request: {e}")
   
    
    @classmethod
    async def process_response(cls, response, response_format):
        """
        Processes the response result before returning it in the specified format.

        Args:
            response (aiohttp.ClientResponse): The response object from the server.
            response_format (str): The format of the response data to return.

        Returns:
            Depends on the specified response format:
            - dict: for 'json' format;
            - str: for 'text' format;
            - bytes: for 'bytes' format.
        """

        if response_format == 'json':
            try:
                return await response.json()
            except aiohttp.ContentTypeError:
                data = await response.text()
                return json.loads(data)
        elif response_format == 'text':
            return await response.text()
        elif response_format == 'bytes':
            return await response.read()
        else:
            raise ValueError("Unsupported response format. Supported formats are 'json', 'text', and 'bytes'.")


from __future__ import annotations
from pydantic import BaseModel, Field, ConfigDict, field_validator
from typing import List, Optional, Dict, Union
import math

ICON_PROPS = {
    11101: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-3.png",  # HP [Base]
    11102: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-3.png",  # HP%
    11103: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-3.png",  # HP [Flat]
    12101: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-4.png",  # ATK [Base]
    12102: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-4.png",  # ATK%
    12103: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-4.png",  # ATK [Flat]
    12201: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-6.png",  # Impact [Base]
    12202: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-6.png",  # Impact%
    13101: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-5.png",  # Def [Base]
    13102: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-5.png",  # Def%
    13103: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-5.png",  # Def [Flat]
    20101: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-7.png",  # Crit Rate [Base]
    20103: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-7.png",  # Crit Rate [Flat]
    21101: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-8.png",  # Crit DMG [Base]
    21103: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-8.png",  # Crit DMG [Flat]
    23101: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-11.png",  # Pen Ratio [Base]
    23103: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-11.png",  # Pen Ratio [Flat]
    23201: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-12.png",  # PEN [Base]
    23203: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-12.png",  # PEN [Flat]
    30501: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-13.png",  # Energy Regen [Base]
    30502: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-13.png",  # Energy Regen%
    30503: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-13.png",  # Energy Regen [Flat]
    31201: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-9.png",  # Anomaly Proficiency [Base]
    31203: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-9.png",  # Anomaly Proficiency [Flat]
    31401: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-10.png",  # Anomaly Mastery [Base]
    31402: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-10.png",  # Anomaly Mastery%
    31403: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-10.png",  # Anomaly Mastery [Flat]
    31501: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-2%20(1).png",  # Physical DMG Bonus [Base]
    31503: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-2%20(1).png",  # Physical DMG Bonus [Flat]
    31601: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-2%20(2).png",  # Fire DMG Bonus [Base]
    31603: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-2%20(2).png",  # Fire DMG Bonus [Flat]
    31701: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-2.png",  # Ice DMG Bonus [Base]
    31703: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-2.png",  # Ice DMG Bonus [Flat]
    31801: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-2%20(3).png",  # Electric DMG Bonus [Base]
    31803: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-2%20(3).png",  # Electric DMG Bonus [Flat]
    31901: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-2%20(4).png",  # Ether DMG Bonus [Base]
    31903: "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/icon/svgexport-2%20(4).png"   # Ether DMG Bonus [Flat]
}


def weapon_main(jsons_data:dict, level:int, asc:int, v:int):
    main_value = jsons_data["weapons_data"][str(v)]["MainStat"]["PropertyValue"]
    wp_level_value = jsons_data["level_format_data"][str(jsons_data["weapons_data"][str(v)]["Rarity"])][str(level)]["value"]
    wp_star_value = jsons_data["star_format_data"][str(jsons_data["weapons_data"][str(v)]["Rarity"])][str(asc)]["main_value"]
    formula = main_value * (1 + wp_level_value / 10000 + wp_star_value / 10000)
    return int(formula)

def weapon_sub(jsons_data:dict, asc:int, v:int):
    main_value = jsons_data["weapons_data"][str(v)]["SecondaryStat"]["PropertyValue"]
    wp_star_value = jsons_data["star_format_data"][str(jsons_data["weapons_data"][str(v)]["Rarity"])][str(asc)]["second_value"]
    formula = main_value * (1 + wp_star_value / 10000)
    return int(formula)

def calculate_stats(jsons_data: dict, level: int, promotion: int, core: int, cid: int, weapon: WeaponData, equipped: List[EquippedList], lang_dict: dict):
   
    promotion = int(promotion)
    avatar_data = jsons_data["avatars_data"].get(str(cid), {})
    if not avatar_data:
        raise ValueError(f"No character data found for CID {cid}")

    base_props = avatar_data.get("BaseProps", {}).copy()
    growth_props = avatar_data.get("GrowthProps", {})
    promotion_props = avatar_data.get("PromotionProps", [])
    core_enhancement_props = avatar_data.get("CoreEnhancementProps", {})

    if promotion - 1 >= len(promotion_props):
        raise ValueError(f"Incorrect promotion {promotion} for CID {cid}")

    promotion_values = promotion_props[promotion - 1]
    core_enhancement_values = core_enhancement_props[core]

    for key, base_value in base_props.items():
        growth_value = (growth_props.get(key, 0) * (level - 1)) / 10000
        promotion_value = promotion_values.get(key, 0)
        core_enhancement_value = core_enhancement_values.get(key, 0)

        base_props[key] = int(base_value) + int(growth_value) + int(promotion_value) + int(core_enhancement_value)
    if "23103" not in base_props:
        base_props["23103"] = 0
    if "23203" not in base_props:
        base_props["23203"] = 0

    main_base_props = base_props.copy()

    if weapon:
        base_props["12101"] += weapon.main.value
        for stat in [weapon.sub]:
            stat_id = str(stat.id)
            if stat_id in base_props:
                base_props[stat_id] += stat.value
            elif str(stat.id - 1) in base_props and stat.id - 1 not in [20101, 21101, 23101]:
                base_props[str(stat.id - 1)] += (stat.value / 10000) * main_base_props[str(stat.id - 1)]
            elif str(stat.id - 2) in base_props:
                base_props[str(stat.id - 2)] += stat.value
            else:
                base_props[stat_id] = stat.value
    sets = []
    for item in equipped:
        equipment = item.equipment
        if not equipment:
            continue
        sets.append(equipment.sets)
        for main_stat in equipment.main:
            stat_id = str(main_stat.id)
            if stat_id in base_props:
                base_props[stat_id] += main_stat.value
            elif str(main_stat.id-2) in base_props:
                base_props[str(main_stat.id-2)] += main_stat.value
            elif str(main_stat.id - 1) in base_props and main_stat.id - 1 not in [20101, 21101, 23101]:
                base_props[str(main_stat.id - 1)] += (main_stat.value / 10000) * main_base_props[str(main_stat.id - 1)]
            else:
                base_props[stat_id] = main_stat.value
        for sub_stat in equipment.sub:
            stat_id = str(sub_stat.id)
            if stat_id in base_props:
                base_props[stat_id] += sub_stat.value
            elif str(sub_stat.id - 1) in base_props and sub_stat.id - 1 not in [20101, 21101, 23101]:
                base_props[str(sub_stat.id - 1)] += (sub_stat.value / 10000) * main_base_props[str(sub_stat.id - 1)]
            elif str(sub_stat.id - 2) in base_props:
                base_props[str(sub_stat.id - 2)] += sub_stat.value
            else:
                base_props[stat_id] = sub_stat.value
                
    main_bonus = []
    for ijk in sets:
        if ijk not in main_bonus:
            cou = sets.count(ijk)
            if cou >= 2:
                main_bonus.append(ijk)
    for mnp in main_bonus:
        i = mnp.bonus
        if i == {}:
            continue
        sid = list(i.keys())[0]
        val = list(i.values())[0]
        if sid in base_props:
            base_props[sid] += val
        elif str(int(sid) - 1) in base_props and str(int(sid) - 1) not in [20101, 21101, 23101]:
            base_props[str(int(sid) - 1)] += (val / 10000) * main_base_props[str(int(sid) - 1)]
        elif str(int(sid) - 2) in base_props:
            base_props[str(int(sid) - 2)] += val
        else:
            base_props[sid] = val

    property_data = jsons_data.get("property_data", {})
    BaseStats = [
        Stat(
            id=int(key),
            key = property_data.get(key, {}).get("Name", "Unknown"),
            name= lang_dict.get(property_data.get(key, {}).get("Name", "Unknown")),
            value= int(value),
            format=property_data.get(key, {}).get("Format", "{}"),
            base = main_base_props.get(key, 0),
            add = int(value) - main_base_props.get(key, 0),
            icon = ICON_PROPS.get(int(key), "")
        )
        for key, value in base_props.items()
    ]
    return BaseStats


class CharacterDataHoYo(BaseModel):
    model_config = ConfigDict(from_attributes=True)
    name: str
    link: str
    id: str
    faction: Optional[Union[str, List[str]]] = None
    type: Optional[str] = None
    element: str


class Stat(BaseModel):
    id: int
    key: str
    name: str
    value: Union[str, int,float]
    icon: str = None
    format: str
    base: int = None
    add: int = None

    
    def get_value(self, base: bool = False, add: bool = False) -> str:
        if base:
            val = self.base
        elif add:
            val = self.add
        else:
            val = self.value

        if self.id in [12101, 13101, 12201, 31401, 31201, 23203]:
            return str(int(val))
        elif self.id == 30501:
            return "{:.1f}".format(math.floor((val/100) * 10) / 10)
        if self.format == "{0:0}":
            return "{:.0f}".format(math.floor(val))
        elif self.format == "{0:0.#}":
            return "{:.1f}".format(math.floor(val * 10) / 10)
        elif self.format == "{0:0.##}":
            return "{:.2f}".format(math.floor(val * 100) / 100)
        elif self.format == "{0:0.#%}":
            return "{:.1f}%".format(math.floor((val/100) * 10) / 10)
        else:
            return val

class Title(BaseModel):
    id: int = None
    name: str = None
    color: Color = None

class Medal(BaseModel):
    id: int = Field(alias="MedalIcon")
    name: str = None
    value: int = Field(alias="Value")
    type: int = Field(alias="MedalType")
    icon: str = None

class ProfileDetail(BaseModel):
    nickname: str = Field(alias="Nickname")
    avatar_id: int = Field(alias="AvatarId")
    icon: str = None
    uid: int = Field(alias="Uid")
    level: int = Field(alias="Level")
    title: Union[Title, int] = Field(alias="Title")
    pfp_id: int = Field(alias="ProfileId")

    @field_validator("title")
    def validate_title(cls, value):
        if isinstance(value, int):
            return Title(id=value)
        return value

class WeaponData(BaseModel):
    id: int = Field(alias="Id")
    uid: int = Field(alias="Uid")
    name: str = None
    rarity: int = None
    level: int = Field(alias="Level")
    cons: int = Field(alias="UpgradeLevel")
    icon: str = None
    professio: str = None
    main: WeaponProps = None
    sub: WeaponProps = None
    asc: int = Field(alias="BreakLevel")

class SkillLevelList(BaseModel):
    level: int = Field(alias="Level")
    index: int = Field(alias="Index")

class SubProperty(BaseModel):
    id: int = Field(alias="PropertyId")
    rolls: int = Field(alias="PropertyLevel")
    icon: str = None
    value: int = Field(alias="PropertyValue")
    format: str = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.icon = ICON_PROPS.get(int(self.id), "")
        self.value *= self.rolls

    def get_value(self) -> str:
        if self.id in [12103, 13103]:
            return str(int(self.value))
        if self.format == "{0:0}":
            return "{:.0f}".format(math.floor(self.value))
        elif self.format == "{0:0.#}":
            return "{:.1f}".format(math.floor(self.value * 10) / 10)
        elif self.format == "{0:0.##}":
            return "{:.2f}".format(math.floor(self.value * 100) / 100)
        elif self.format == "{0:0.#%}":
            return "{:.1f}%".format(math.floor((self.value/100) * 10) / 10)
        else:
            return self.value

class MainProperty(BaseModel):
    id: int = Field(alias="PropertyId")
    icon: str = None
    value: int = Field(alias="PropertyValue")
    format: str = None
    

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.icon = ICON_PROPS.get(int(self.id), "")

    def get_value(self) -> str:
        if self.id in [12103, 13103]:
            return str(int(self.value))
        if self.format == "{0:0}":
            return "{:.0f}".format(math.floor(self.value))
        elif self.format == "{0:0.#}":
            return "{:.1f}".format(math.floor(self.value * 10) / 10)
        elif self.format == "{0:0.##}":
            return "{:.2f}".format(math.floor(self.value * 100) / 100)
        elif self.format == "{0:0.#%}":
            return "{:.1f}%".format(math.floor((self.value/100) * 10) / 10)
        else:
            return self.value 

class Equipment(BaseModel):
    id: int = Field(alias="Id")
    uid: int = Field(alias="Uid")
    rarity: int = 2
    icon: str = None
    level: int = Field(alias="Level")
    break_level: int = Field(alias="BreakLevel")
    max_level: int = 9
    avalibale: bool = Field(alias="IsAvailable")
    locked: bool = Field(alias="IsLocked")
    trash: bool = Field(alias="IsTrash")
    main: List[MainProperty] = Field(alias="MainPropertyList")
    sub: List[SubProperty] = Field(alias="RandomPropertyList")
    sets: Sets = None

class EquippedList(BaseModel):
    slot: int = Field(alias="Slot")
    equipment: Equipment = Field(alias="Equipment")

class CharterShowCase(BaseModel):
    id: int
    name: str
    icon: CharterIcon
    element: Element
    rarity: int
    color: Color
    level: int
    max_level: int


class Character(BaseModel):
    id: int = Field(alias="Id")
    name: str = None
    icon: CharterIcon = None
    element: Element = None
    faction: str = None
    profession: Profession = None
    rarity: int = None
    color: dict = None
    level: int = Field(alias="Level")
    max_level: int = Field(alias="PromotionLevel")
    skills: List[SkillLevelList] = Field(alias="SkillLevelList")
    core_skill: int = Field(alias="CoreSkillEnhancement")
    cinema: List[bool] = Field(alias="TalentToggleList")
    const: int = Field(alias="TalentLevel")
    weapon: Optional[WeaponData]  = Field(None, alias="Weapon")
    equippe: List[EquippedList] = Field(alias="EquippedList")
    stats: List[Stat] = None
    reward: Optional[List[int]] = Field(alias="ClaimedRewardList")
    
    model_config = ConfigDict(populate_by_name=True)

class CharacterData(BaseModel):
    characters: List[Character] = Field(alias="AvatarList")

class PlayerData(BaseModel):
    medal_list: List[Medal] = Field(alias="MedalList")
    profile: ProfileDetail = Field(alias="ProfileDetail")
    desc: str = Field(alias="Desc")
    characters: list = []

class CharterIcon(BaseModel):
    icon: str
    circle_icon: str

class Element(BaseModel):
    name: str = ""
    icon: str = ""

class Profession(BaseModel):
    name: str
    icon: str

class Color(BaseModel):
    accent: str
    mindscape: str

class Sets(BaseModel):
    id: int
    name: str
    icon: str = None
    bonus: Dict[str, int]

class WeaponProps(BaseModel):
    id: int
    value: int
    icon: str = None
    format: str = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


    def get_value(self) -> str:
        if self.id in [12103, 13103]:
            return str(int(self.value))
        if self.format == "{0:0}":
            return "{:.0f}".format(math.floor(self.value))
        elif self.format == "{0:0.#}":
            return "{:.1f}".format(math.floor(self.value * 10) / 10)
        elif self.format == "{0:0.##}":
            return "{:.2f}".format(math.floor(self.value * 100) / 100)
        elif self.format == "{0:0.#%}":
            return "{:.1f}%".format(math.floor((self.value/100) * 10) / 10)
        else:
            return self.value


class ZenkaApi(BaseModel):
    player: Optional[PlayerData] = Field(alias="SocialDetail")
    character_info: CharacterData = Field(alias="ShowcaseDetail")

    def __init__(self, jsons_data: Dict, lang: str, **data):
        super().__init__(**data)
        data_charter = jsons_data.get("avatars_data")
        data_lang = jsons_data.get("locs_data")
        data_lang = data_lang.get(lang)
        hoyolink_data = jsons_data.get("hoyolink_data")
        weapon_data = jsons_data.get("weapons_data")
        equipments_data = jsons_data.get("equipments_data")
        property_data = jsons_data.get("property_data", {})
        medal_data = jsons_data.get("medals_data", {})
        titles_data = jsons_data.get("titles_data", {})
        pfps_data = jsons_data.get("pfps_data", {})

        for charters in self.character_info.characters:
            data_charters = data_charter.get(str(charters.id))
            hoyolink_datas = hoyolink_data.get(str(charters.id))
            charters.name = data_lang.get(data_charters.get("Name"))
            charters.icon = CharterIcon(icon = data_charters.get("Image"), circle_icon = data_charters.get("CircleIcon"))
            charters.element = Element(name = data_charters.get("ElementTypes")[0], icon=hoyolink_datas.get("element"))
            charters.profession = Profession(name= data_charters.get("ProfessionType"), icon=hoyolink_datas.get("type"))
            charters.faction = hoyolink_datas.get("faction")
            charters.rarity = data_charters.get("Rarity")
            charters.color = Color(accent=data_charters.get("Colors").get("Accent"), mindscape = data_charters.get("Colors").get("Mindscape"))
            
            if charters.weapon:
                weapon_datas = weapon_data.get(str(charters.weapon.id)) 
                charters.weapon.name = data_lang.get(weapon_datas.get("ItemName"))
                charters.weapon.icon = weapon_datas.get("ImagePath")
                charters.weapon.professio = weapon_datas.get("ProfessionType")
                charters.weapon.rarity = weapon_datas.get("Rarity")
                main_id = int(weapon_datas.get("MainStat").get("PropertyId"))
                charters.weapon.main = WeaponProps(id = main_id, value = weapon_main(jsons_data, charters.weapon.level, charters.weapon.asc,charters.weapon.id), icon = ICON_PROPS.get(main_id, ""))
            
                sub_id = int(weapon_datas.get("SecondaryStat").get("PropertyId"))
                charters.weapon.sub = WeaponProps(id = sub_id, value = weapon_sub(jsons_data, charters.weapon.asc,charters.weapon.id), icon = ICON_PROPS.get(sub_id, ""), format=property_data.get(str(sub_id), {}).get("Format"))

            self.player.characters.append(
                CharterShowCase(
                    id = charters.id,
                    name = charters.name,
                    rarity=charters.rarity,
                    icon= charters.icon,
                    element = charters.element,
                    color = charters.color,
                    level = charters.level,
                    max_level = charters.max_level
                )
            )
            for key in charters.equippe:
                sets_id = equipments_data["Items"].get(str(key.equipment.id)).get("SuitId")
                key.equipment.icon = equipments_data["Suits"].get(str(sets_id)).get("Icon")
                key.equipment.rarity = equipments_data["Items"].get(str(key.equipment.id)).get("Rarity")
                key.equipment.main[0].value *= (1 + jsons_data["discs_format_data"][str(key.equipment.rarity)][str(key.equipment.level)] / 10000)
                key.equipment.sets = Sets(
                    id = sets_id,
                    name = data_lang.get(equipments_data["Suits"].get(str(sets_id)).get("Name")),
                    icon = key.equipment.icon,
                    bonus = equipments_data["Suits"].get(str(sets_id)).get("SetBonusProps")
                )


            charters.stats = calculate_stats(jsons_data, charters.level, charters.max_level / 10, charters.core_skill, charters.id, charters.weapon, charters.equippe, data_lang)

            for key in charters.equippe:
                key.equipment.main[0].format = property_data.get(str(key.equipment.main[0].id), {}).get("Format", "{}")
                for subs in key.equipment.sub:
                    subs.format = property_data.get(str(subs.id), {}).get("Format", "{}")

        for medal in self.player.medal_list:
            if medal.id == 0:
                continue
            
            medal.name = data_lang.get(medal_data.get(str(medal.id)).get("Name"))
            medal.icon = medal_data.get(str(medal.id)).get("Icon")
        
        title = titles_data.get(str(self.player.profile.title.id))
        self.player.profile.icon = pfps_data.get(str(self.player.profile.pfp_id)).get("Icon")
        if title:
            self.player.profile.title.name =  data_lang.get(title.get("TitleText"))
            self.player.profile.title.color = Color(accent= title.get("ColorA"), mindscape= title.get("ColorB"))
        else:
            self.player.profile.title = None


from .zenka import *
from .src.model.assets import *
from .src.tools.error import *
from .src.tools.api import *
from .src.model.generator import *
from .src.model import assets
from .src.model.base import *


# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.

from PIL import Image,ImageFilter
from more_itertools import chunked

import numpy as np

from .color_control import light_level, get_light_pixel_color, _get_dark_pixel_color

async def get_average_color(image):
    if image.mode != 'RGBA':
        image = image.convert('RGBA')
    
    channels = image.split()
    
    return (
        round(np.average(channels[0], weights=channels[-1])),
        round(np.average(channels[1], weights=channels[-1])),
        round(np.average(channels[2], weights=channels[-1])),
    )


async def get_dominant_colors(
    image,
    number,
    *,
    dither=Image.Quantize.FASTOCTREE,
    common=True,
):
    if image.mode != 'RGB':
        if image.mode != 'RGBA':
            image = image.convert('RGBA')
        
        if not common:
            width = image.width
            height = image.height
            
            image = Image.fromarray(np.array([np.repeat(
                np.reshape(image.convert('RGB'), (width * height, 3)),
                np.reshape(image.split()[-1], width * height),
                0,
            )]), 'RGB')
    
    if image.mode == 'RGBA':
        if dither == Image.Quantize.FASTOCTREE:
            simple_image = image.copy()
            simple_image.putalpha(255)
        else:
            simple_image = image.convert('RGB')
    else:
        simple_image = image
    
    reduced = simple_image.quantize(dither=dither, colors=number)
    
    palette = [*chunked(reduced.getpalette(), 3)]
    
    if common and image.mode == 'RGBA':
        alpha = np.array(image.split()[-1])
        
        colors = sorted((
            (
                np.sum(alpha * reduced.point([0] * i + [1] + [0] * (255 - i))),
                tuple(palette[i]),
            )
            for _, i in reduced.getcolors()
        ), reverse=True)
    else:
        colors = [
            (n, tuple(palette[i]))
            for n, i in sorted(reduced.getcolors(), reverse=True)
        ]
    
    return tuple(colors)


async def get_distance_alpha(image, converter=(lambda x: x)):
    width = image.width
    height = image.height
    
    radius = np.hypot(1, 1)
    
    return Image.fromarray(np.fromfunction(
        lambda y, x: np.uint8(255 * converter(np.hypot(
            2 * x / (width - 1) - 1,
            2 * y / (height - 1) - 1,
        ) / radius)),
        (height, width),
    ), 'L')


async def get_background_alpha(image):
    return await get_distance_alpha(
        image,
        lambda x: x * np.sin(x * np.pi / 2),
    )


async def get_foreground_alpha(image):
    return await get_distance_alpha(
        image,
        lambda x: 1 - x * np.sin(x * np.pi / 2),
    )

async def get_colors(image,number,*,common=False,radius=1,quality=None):
    if quality is not None:
        image = image.copy()
        image.thumbnail((quality, quality), 0)
    
    if radius > 1:
        image = image.filter(ImageFilter.BoxBlur(radius))
    
    filtered_image = image.convert('RGB')
    
    if image.mode != 'RGBA':
        filtered_image.putalpha(await get_background_alpha(image))
    else:
        filtered_image.putalpha(Image.fromarray(np.uint8(
            np.uint16(await get_background_alpha(image))
            * image.split()[-1]
            / 255
        ), 'L'))
    
    color_palette = await get_dominant_colors(filtered_image, number, common=common)
    color_palette = color_palette[0][1]
    ll = await light_level(color_palette)
    if ll < 0.15:
        color_palette = await get_light_pixel_color(color_palette)
    elif ll > 0.80:
        color_palette = await _get_dark_pixel_color(color_palette)
        
        
    return color_palette


# Copyright 2025 DEViantUa <t.me/deviant_ua>
# All rights reserved.

from PIL import Image
import os
import threading
from pathlib import Path
from io import BytesIO

from .http import AioSession
from .cache import Cache

lock = threading.Lock()

_caches = Cache.get_cache()

assets = Path(__file__).parent.parent / 'assets'

_BASE_URL = 'https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/enka/'

font = str(assets / 'font' / 'impact.ttf')


def determine_font_path_automatically(font_file = 'Times New Roman.ttf'):
    font_dirs = [
        '/usr/share/fonts',          
        '/usr/local/share/fonts',    
        '/Library/Fonts',           
    ]
        
    for font_dir in font_dirs:
        font_path = os.path.join(font_dir, font_file)
        if os.path.isfile(font_path):
            return font_path

    return None

async def change_font(font_path = None):
    global font
    if font_path is None:
        font = str(assets / 'font' / 'impact.ttf')
    else:
        font_path = os.path.abspath(font_path)
        if os.path.isfile(font_path):
            font = font_path 
        else:
            font_path = determine_font_path_automatically(font_path)
            if font_path is None:
                font = str(assets / 'font' / 'impact.ttf')


total_style = {
    'g_five': 'stars/five.png',
    'g_four': 'stars/four.png',
    'g_three': 'stars/three.png',
    'g_two': 'stars/two.png',
    'g_one': 'stars/one.png',
    "rank_a": 'rank_icon/a.png',
    "rank_b": 'rank_icon/b.png',
    "rank_s": 'rank_icon/s.png',

    "rank_w_a": 'rank_icon/w_a.png',
    "rank_w_b": 'rank_icon/w_b.png',
    "rank_w_s": 'rank_icon/w_s.png',

    "const_0": 'const/const_0.png',
    "const_1": 'const/const_1.png',
    "const_2": 'const/const_2.png',
    "const_3": 'const/const_3.png',
    "const_4": 'const/const_4.png',
    "const_5": 'const/const_5.png',
    "const_6": 'const/const_6.png',

    "skill_1_s": 'skill_icon/IconRoleSkillKeyNormal.png',
    "skill_2_s": 'skill_icon/IconRoleSkillKeyEvade.png',
    "skill_3_s": 'skill_icon/IconRoleSkillKeySwitch.png',
    "skill_4_s": 'skill_icon/IconRoleSkillKeySpecialV2.png',
    "skill_5_s": 'skill_icon/IconRoleSkillKeyUltimateV2.png',
    "skill_6_s": 'skill_icon/IconRoleSkillKe.png',

    "core_skill_a": "core_skill/svgexport-15.png",
    "core_skill_b": "core_skill/svgexport-16.png",
    "core_skill_c": "core_skill/svgexport-17.png",
    "core_skill_d": "core_skill/svgexport-18.png",
    "core_skill_e": "core_skill/svgexport-19.png",
    "core_skill_f": "core_skill/svgexport-20.png",

    "conts_0_lvl": "const/const_0.png",
    "conts_1_lvl": "const/const_1.png",
    "conts_2_lvl": "const/const_2.png",
    "conts_3_lvl": "const/const_3.png",
    "conts_4_lvl": "const/const_4.png",
    "conts_5_lvl": "const/const_5.png",
    "conts_6_lvl": "const/const_6.png"
    
}


card_style = {
    'background': 'background/background.png',
    'cadr': 'background/cadr.png',
    'charter_bg': 'background/charter_bg.png',
    'dark_frame_info': 'background/dark_frame_info.png',
    'frame_info': 'background/frame_info.png',
    'frame': 'background/frame.png',
    'mask': 'background/mask.png',
    'shadow': 'background/shadow.png',
    'text_frame': 'background/text_frame.png',


    'frame_background_black': 'relict/frame_background_black.png',
    'frame_background': 'relict/frame_background.png',
    'frame_relict': 'relict/frame.png',
    'frame_wb_background': 'relict/frame_wb_background.png',
    'lvl': 'relict/lvl.png',
    'rank_leve': "relict/rank_level.png",
    'sets_background': 'relict/sets_background.png',
    'sets_collor': 'relict/sets_collor.png',
    'sets_count': 'relict/sets_count.png',
    'stats': 'relict/stats.png',


    'stats_background': 'stats/background.png',
    'skill_count': 'stats/skill_count.png',
    'value': 'stats/value.png',

    'weapon_background': 'weapon/background.png',
    'levl': 'weapon/levl.png',
    'up': 'weapon/up.png',
    
}

profile_style = {
    'main_background': 'profile_style_one/main_background.png',
    'background': 'profile_style_one/background.png',
    'cinema_frame': 'profile_style_one/cinema_frame.png',
    'cinema': 'profile_style_one/cinema.png',
    'frame_avatar': 'profile_style_one/frame_avatar.png',
    'frame_icon': 'profile_style_one/frame_icon.png',

    'lvl_b': 'profile_style_one/lvl_b.png',
    'lvl_c': 'profile_style_one/lvl_c.png',
    'lvl_w': 'profile_style_one/lvl_w.png',

    'mask': 'profile_style_one/mask.png',
    'shadow_charter': 'profile_style_one/shadow_charter.png',
    'texture_avatar': 'profile_style_one/texture_avatar.png',


    'uid_b': 'profile_style_one/uid_b.png',
    'uid_c': 'profile_style_one/uid_c.png',
    'uid_w': 'profile_style_one/uid_w.png',
    'medal':'profile_style_one/medal.png'
    
}


card_style_team = {
    'background':'style_teams/background.png',
    'background_mask':'style_teams/background_mask.png',
    'text_frame':'style_teams/text_frame.png',

    'lvl_disk':'style_teams/lvl_disk.png',
    'frame_disk_lvl':'style_teams/frame_disk_lvl.png',
    'disk_bg':'style_teams/disk_bg.png',


    'main_bg_disk_stat':'style_teams/main_bg_disk_stat.png',
    'sub_bg_disk_stat':'style_teams/sub_bg_disk_stat.png',
    'texture_disk_main':'style_teams/texture_disk_main.png',
    'rank_leve': "relict/rank_level.png",

    'full_bg':'style_teams/full_bg.png',
    'full_bg_b':'style_teams/full_bg_dark.png',

    'info_bg':'style_teams/info_bg.png',
    'stats_bg':'style_teams/stats_bg.png',

    'weapon_bg':'style_teams/weapon_bg.png',
    'levl': 'weapon/levl.png',
    'up': 'weapon/up.png',
    
    'skill_bg':'style_teams/skill_bg.png',
    'skill_frame':'style_teams/skill_frame.png',
    'skill_lvl':'style_teams/skill_lvl.png',
}

class ImageCache:
    
    _assets_download = False
    _mapping = {}
            
    @classmethod
    async def set_assets_download(cls, download = False):
        cls._assets_download = download
    
    @classmethod
    def set_mapping(cls,style: int = 1) -> None:
        if style == 1:
            cls._mapping = card_style
        elif style == 2:
            cls._mapping = profile_style
        elif style == 3:
            cls._mapping = card_style_team
        
    @classmethod
    async def _load_image(cls, name) -> Image.Image:
        
        try:
            image = _caches[name]
        except KeyError:
            try:
                _caches[name] = image = Image.open(assets / name)
                return _caches[name]
            except Exception as e:
                pass
        
        try:
            _caches[name] = image = Image.open(assets / name)
            return _caches[name]
        except Exception as e:
            pass
        
        url = _BASE_URL + name
        if url in _caches:
            return _caches[name]
        else:
            image_data = await AioSession.get(url, response_format= "bytes")
            image = Image.open(BytesIO(image_data))
            _caches[name] = image
        
        if cls._assets_download:
            file_path = assets / name
            file_path.parent.mkdir(parents=True, exist_ok=True)
            image.save(str(assets / name))
        
        return image

    async def __getattr__(cls, name) -> Image.Image:
        if name in cls._mapping:
            return await cls._load_image(cls._mapping[name])
        else:
            if name in total_style:
                return await cls._load_image(total_style[name]) 
            else:
                raise AttributeError(f"'{cls.__class__.__name__}' object has no attribute '{name}'")


# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.
import colorsys
from typing import Union
from PIL import Image
from .. import cache 

_caches = cache.Cache.get_cache()

def is_white_visible(background_color: tuple, threshold: float = 4.5) -> bool:
    def relative_luminance(rgb):
        def transform(c):
            c /= 255
            return c / 12.92 if c <= 0.03928 else ((c + 0.055) / 1.055) ** 2.4
        
        r, g, b = rgb
        return 0.2126 * transform(r) + 0.7152 * transform(g) + 0.0722 * transform(b)

    # Яркость белого цвета
    luminance_white = relative_luminance((255, 255, 255))
    # Яркость фона
    luminance_bg = relative_luminance(background_color)
    
    # Рассчет коэффициента контраста
    contrast_ratio = (luminance_white + 0.05) / (luminance_bg + 0.05)

    return contrast_ratio >= threshold

async def apply_opacity(image, opacity=0.2):
    result_image = image.copy()
    alpha = result_image.split()[3]
    alpha = alpha.point(lambda p: int(p * opacity))
    result_image.putalpha(alpha)

    return result_image

def brightness(color):
    return 0.299 * color[0] + 0.587 * color[1] + 0.114 * color[2]

async def light_levell(pixel_color: tuple):
    cache_key = pixel_color
    if cache_key in _caches:
        return _caches[cache_key]

    luminance = brightness(pixel_color)
    _caches[cache_key] = luminance
    return luminance

async def is_white_text_readable(pixel_color: tuple, threshold=200):
    bg_brightness = await light_levell(pixel_color)
    white_brightness = 255

    contrast = abs(white_brightness - bg_brightness)
    return contrast >= threshold 


async def light_level(pixel_color: tuple):
    cache_key = pixel_color
    if cache_key in _caches:
        return _caches[cache_key]
    
    h, l, s = colorsys.rgb_to_hls(*(x / 255 for x in pixel_color[:3])) 
    _caches[cache_key] = l
    return l

def color_distance(color1, color2):
    return sum((a - b) ** 2 for a, b in zip(color1, color2)) ** 0.5

async def replace_color(image, old_color, new_color, radius=100):
    image = image.convert("RGBA")
    pixels = image.load()
    width, height = image.size

    for y in range(height):
        for x in range(width):
            current_color = pixels[x, y][:3]
            if color_distance(current_color, old_color) <= radius:
                pixels[x, y] = (*new_color, pixels[x, y][3])
    
    return image


async def recolor_image(image, target_color, light = False) -> Image.Image:
    if light:
        ll = await light_level(target_color)
        if ll < 45:
            target_color = await get_light_pixel_color(target_color,up = True)
    if image.mode != 'RGBA':
        image = image.convert('RGBA')

    image = image.copy()

    pixels = image.load()
    for i in range(image.width):
        for j in range(image.height):
            r, g, b, a = pixels[i, j]
            if a != 0:
                pixels[i, j] = target_color + (a,)
    if light:
        return image, target_color
    return image

async def get_light_pixel_color(pixel_color, up = False):
    h, l, s = colorsys.rgb_to_hls(*(x / 255 for x in pixel_color[:3]))
    if up:
        l = min(max(0.6, l), 0.9)
    else:
        l = min(max(0.3, l), 0.8)
    return tuple(round(x * 255) for x in colorsys.hls_to_rgb(h, l, s))
  
async def _get_dark_pixel_color(pixel_color):
    h, l, s = colorsys.rgb_to_hls(*(x / 255 for x in pixel_color[:3]))
    l = min(max(0.8, l), 0.2)
    a = tuple(round(x * 255) for x in colorsys.hls_to_rgb(h, l, s))
    
    return  a






# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.

import os
import re
import json

from PIL import Image
from io import BytesIO
from ..git import assets
from .. import cache, http

_caches = cache.Cache.get_cache()
_boost_speed = False


async def resize_image(image, scale):
    width, height = image.size
    new_width = int(width * scale)
    new_height = int(height * scale)
    resized_image = image.resize((new_width, new_height))
    return resized_image


async def get_center_size(size, file_name):
    background_image = Image.new('RGBA', size, color=(0, 0, 0, 0))
    foreground_image = file_name.convert("RGBA")

    scale = max(size[0] / foreground_image.size[0], size[1] / foreground_image.size[1])
    foreground_image = foreground_image.resize((int(foreground_image.size[0] * scale), int(foreground_image.size[1] * scale)))

    background_size = background_image.size
    foreground_size = foreground_image.size

    x = background_size[0] // 2 - foreground_size[0] // 2

    if foreground_size[1] > background_size[1]:
        y_offset = max(int(0.3 * (foreground_size[1] - background_size[1])), int(0.5 * (-foreground_size[1])))
        y = -y_offset
    else:
        y = background_size[1] // 2 - foreground_size[1] // 2

    background_image.alpha_composite(foreground_image, (x, y))

    return background_image


async def get_center_scale(size, file_name):
    
    background_image = Image.new('RGBA', size, color=(0, 0, 0, 0))
    foreground_image = file_name.convert("RGBA")
    
    scale = 0.65
    foreground_image = await resize_image(foreground_image, scale)

    background_size = background_image.size
    foreground_size = foreground_image.size

    x = background_size[0] // 2 - foreground_size[0] // 2
    y = background_size[1] // 2 - foreground_size[1] // 2

    background_image.alpha_composite(foreground_image, (x, y))

    return background_image


async def get_image_from_boost_speed(link):
    path = f"/boost_speed/{link.split('master')[1]}"
    full_path = os.path.join(assets, path.lstrip('/'))

    if os.path.exists(full_path):
        try:
            return Image.open(full_path).convert("RGBA")
        except Exception as e:
            raise IOError(f"Error reading image file: {e}")
    else:
        return None

async def download_image(link, headers=None):
    try:
        image = await http.AioSession.get(link, headers=headers, response_format="bytes")
        return image
    except:
        raise TypeError(f"Error Download image: {link}")

async def save_image(image, full_path):
    directory = os.path.dirname(full_path)
    if not os.path.exists(directory):
        os.makedirs(directory)
    
    with open(full_path, 'wb') as f:
        f.write(image)

async def open_image(image_data):
    try:
        return Image.open(BytesIO(image_data)).convert("RGBA")
    except Exception as e:
        raise TypeError(f"Error Open image: {image_data}")

async def get_download_img(link, size=None, thumbnail_size=None, gif=False):
    cache_key = json.dumps((link, size, thumbnail_size), sort_keys=True)
    image_boost = None
    
    if _boost_speed:
        if "StarRailRes" in link:
            image_boost = await get_image_from_boost_speed(link)

    if not gif:
        if cache_key in _caches:
            return _caches[cache_key]
    
    
    
    if image_boost is None:
        headers_p = None
    
        if "pximg" in link:
            headers_p = {
                "referer": "https://www.pixiv.net/",
            }
        image = await download_image(link, headers_p)

        if _boost_speed:
            if "StarRailRes" in link:
                full_path = os.path.join(assets, f"/boost_speed/{link.split('master')[1]}".lstrip('/'))
                await save_image(image, full_path)
                      
        image = await open_image(image)
    else:
        image = image_boost

    if gif:
        return image
    
    try:
        if size:
            image = image.resize(size)
            _caches[cache_key] = image
            return image
        elif thumbnail_size:
            image.thumbnail(thumbnail_size)
            _caches[cache_key] = image
            return image
        else:
            _caches[cache_key] = image
            return image
    except Exception as e:
        raise TypeError(f"Error setting image: {link}")
    
async def crop_image(img):
    width, height = img.size
    target_pixel_x = 275
    target_pixel_y = height // 2
    crop_size = 8
    left = max(0, target_pixel_x - crop_size // 2)
    right = min(width, target_pixel_x + crop_size // 2 + 1)
    cropped_img = img.crop((left, 0, right, height))
    
    return cropped_img

async def get_user_image(img):
    if type(img) != str:
        img = img
    elif type(img) == str:
        linkImg = re.search("(?P<url>https?://[^\\s]+)", img)
        if linkImg:
            try:
                if "gif" in linkImg.group():
                    img = await get_download_img(linkImg.group(), gif = True)
                    return img
                else:
                    img = await get_download_img(linkImg.group())
            except:
                return None
        else:
            img = Image.open(img)
    else:
        return None
    return img.convert("RGBA")



import datetime
import asyncio
import os
from PIL import Image
from typing import Dict, List, Union, Optional, Tuple
from .src.model.base import Config, Lang, ErrorText, TranslationLang
from .src.model.generator import ZenkaGenerator, ZenkaGeneratorTeams, ZenkaGeneratorTeams
from .src.tools import cache, error, http, git, options, api
from .src.generator import style_one, style_one_profile, style_two

def save_file(id: int, card: Image.Image):
    data = datetime.datetime.now().strftime("%d_%m_%Y %H_%M")
    folder = os.path.dirname(f"ZCard/{id}_{data}.png")
    if not os.path.exists(folder):
        os.makedirs(folder, exist_ok=True)
    
    card.save(f"ZCard/{id}_{data}.png")

class Client:
    def __init__(
        self, lang: Union[Lang,str] = Lang.EN.value,
        character_art: Optional[Dict[Union[int,str], str]] = None,
        character_id: Optional[List[int]] = None,
        config: Config = None
    ):
        self.lang = lang.value if isinstance(lang, Lang) else lang
        self.character_art = character_art if character_art is not None else {}
        self.character_id = character_id if character_id is not None else []
        self.config = config if config is not None else Config()


    async def __aenter__(self):
        cache.Cache.get_cache(maxsize = self.config.cache.maxsize, ttl = self.config.cache.ttl)
        await http.AioSession.enter(self.config.proxy)
        await git.ImageCache.set_assets_download(self.config.asset_save)
        await git.change_font(font_path = self.config.font)
        await api.DataManager().collect_data(self.config.asset_save)
        if not isinstance(self.character_art, dict):
            raise error.ZZZError(ErrorText.charterArt)
    
        if not Lang.is_supported(self.lang):
            raise error.ZZZError(ErrorText().format(self.lang))
        
        self.translateLang = TranslationLang().lang(self.lang)
        api.lang = self.lang
        self.config.color = await options.get_color_user(self.config.color)

        return self
    
    async def __aexit__(self, exc_type, exc, tb):
        await http.AioSession.exit(exc_type, exc, tb)

    async def get_api(self, uid: Union[str,int], original: bool = False) -> api.ZenkaApi:
        return  await api.fetch_user(uid, original)
    
    async def update_asset(self) -> None:
        await api.DataManager().update_data()

    async def profile(self, uid: int, color: Tuple[int,int,int,int] = None) -> ZenkaGenerator:
        result = ZenkaGenerator()
        data =  await api.fetch_user(uid)
        for key in data.character_info.characters:           
            result.charter_name.append(key.name)
            result.charter_id.append(key.id)

        result.cards.append(await style_one_profile.StyleOneProfile(key, data.player, self.translateLang, color = color, hide = self.config.hide_uid).start())
        result.player = data.player
        
        if self.config.save:
            for key in result.cards:
                save_file(uid, key.card)
        
        return result
    
    async def card(self, uid: int) -> ZenkaGenerator:
        result = ZenkaGenerator()
        data =  await api.fetch_user(uid)
        task = []
        for key in data.character_info.characters:
            result.charter_name.append(key.name)
            result.charter_id.append(key.id)

            if self.character_id:
                if not key.id in self.character_id:
                    continue 
        
            task.append(style_one.StyleOne(key, data.player, self.translateLang, art = self.character_art.get(key.id), color = self.config.color.get(key.id), hide = self.config.hide_uid, crop = self.config.crop).start())
        
        result.player = data.player
        result.cards = await asyncio.gather(*task)


        if self.config.save:
            for key in result.cards:
                save_file(key.id, key.card)


        return result
    
    async def teams(self, uid: int) -> ZenkaGeneratorTeams:
        result = ZenkaGeneratorTeams()
        data =  await api.fetch_user(uid)
        task = []

        for key in data.character_info.characters:
            result.charter_name.append(key.name)
            result.charter_id.append(key.id)        
            task.append(style_two.StyleTwo(key, data.player, self.translateLang, art = self.character_art.get(key.id), color = self.config.color.get(key.id), hide = self.config.hide_uid, crop = self.config.crop).start())
        
        result.player = data.player
        result.cards = await asyncio.gather(*task)


        if self.config.save:
            for key in result.cards:
                save_file(key.id, key.card)
                
        return result





from typing_extensions import Self
from pydantic import BaseModel, Field, model_validator
from typing import Dict, Tuple, Optional
from enum import Enum


URL_ASSETS = "https://enka.network"
DEFAULT_AVATAR = "https://raw.githubusercontent.com/DEViantUA/ZZZeroCardData/refs/heads/main/default_icon.png"

class HoyoAPIHeaders:
    HEADERS = {
        "accept": "application/json, text/plain, */*",
        "accept-encoding": "gzip, deflate, br, zstd",
        "accept-language": "en-US,en;q=0.9",
        "origin": "https://wiki.hoyolab.com",
        "referer": "https://wiki.hoyolab.com/",
        "sec-fetch-dest": "empty",
        "sec-fetch-mode": "cors",
        "sec-fetch-site": "same-site",
        "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/133.0.0.0 Safari/537.36 Edg/133.0.0.0",
        "x-rpc-language": "en-us",
        "x-rpc-wiki_app": "zzz"
    }

class ElementType(Enum):
    ELEC = "Electric"
    PHYSICS = "Physical"
    ICE = "Ice"
    FIRE = "Fire"
    ETHER = "Ether"

    @classmethod
    def convert(cls, value: str) -> str:
        return cls.__members__.get(value.upper(), "Unknown").value

class Lang(Enum):
    CHT = "cht"
    CN = "cn"
    DE = "de"
    EN = "en"
    ES = "es"
    FR = "fr"
    ID = "id"
    JP = "jp"
    KR = "kr"
    PT = "pt"
    RU = "ru"
    TH = "th"
    VI = "vi"

    @classmethod
    def is_supported(cls, lang: str) -> bool:
        return lang in cls._value2member_map_

class Translator(BaseModel):
    lvl: str
    AR: str
    WL: str
    AC: str
    AB: str


class TranslationLang(BaseModel):
    translations: Dict[str, Translator] = {
        "en": Translator(lvl="LVL", AR="AR", WL="WL", AC="Achievements", AB="Abyss"),
        "ru": Translator(lvl="Уровень", AR="РП", WL="УМ", AC="Достижения", AB="Бездна"),
        "ua": Translator(lvl="Рівень", AR="РП", WL="РС", AC="Досягнення", AB="Безодня"),
        "vi": Translator(lvl="Cấp độ", AR="AR", WL="WL", AC="Thành tích", AB="Vực thẳm"),
        "th": Translator(lvl="ระดับ", AR="AR", WL="WL", AC="ความสำเร็จ", AB="Abyss"),
        "pt": Translator(lvl="Nível", AR="AR", WL="WL", AC="Conquistas", AB="Abismo"),
        "kr": Translator(lvl="레벨", AR="AR", WL="WL", AC="업적", AB="어비스"),
        "jp": Translator(lvl="レベル", AR="AR", WL="WL", AC="アチーブメント", AB="アビス"),
        "zh": Translator(lvl="等级", AR="AR", WL="WL", AC="成就总数", AB="深境螺旋"),
        "cn": Translator(lvl="等级", AR="AR", WL="WL", AC="成就总数", AB="深境螺旋"),
        "id": Translator(lvl="Level", AR="AR", WL="WL", AC="Prestasi", AB="Abyss"),
        "fr": Translator(lvl="Niveau", AR="AR", WL="WL", AC="Réalisations", AB="Abîme"),
        "es": Translator(lvl="Nivel", AR="AR", WL="WL", AC="Logros", AB="Abismo"),
        "de": Translator(lvl="Level", AR="AR", WL="WL", AC="Erfolge", AB="Abyss"),
        "chs": Translator(lvl="等级", AR="AR", WL="WL", AC="成就总数", AB="深境螺旋"),
        "cht": Translator(lvl="等級", AR="AR", WL="WL", AC="成就總數", AB="深境螺旋"),
    }

    def lang(self, lang_code: str) -> Translator:
        return self.translations.get(lang_code, self.translations["en"])
    


class CacheConfig(BaseModel):
    maxsize: int = 150
    ttl: int = 300

class Config(BaseModel):
    font: Optional[str] = None
    save: bool = False
    hide_uid: bool = False
    asset_save: bool = False
    crop: int = 0
    cache: CacheConfig = CacheConfig()
    proxy: Optional[str] = None

    color: Dict[int, Tuple[int, int, int, int]] = Field(default_factory=dict)


    @model_validator(mode='after')
    def check_crop(self) -> Self:
        if self.crop > 120:
            raise ValueError('crop cannot be greater than 50')
        return self

class ErrorText(BaseModel):
    lang: Dict[int, str] = {"code": 1, "text": "This language key [{lang}] is not supported"}
    charterArt: Dict[int, str] = {"code": 2, "text": "The 'character_art' parameter must be of type Dict[Union[int, str], str], "
                "where the key is an integer or string (character ID), "
                "and the value is a string (URL)."
                }
    api: Dict[int, str] = {"code": 3, "text": "{text}"}

    def format(self, **kwargs) -> dict:
        return {"code": self.lang["code"], "text": self.lang["text"].format(**kwargs)}
    
    def format_api(self, **kwargs) -> dict:
        return {"code": self.lang["code"], "text": self.api["text"].format(**kwargs)}




from typing import Dict, Tuple, Union, Optional


async def get_color_user(color: Dict[Union[int, str], Tuple[int, int, int]]) -> Optional[Dict[Union[int, str], Tuple[int, int, int]]]:
    processed_dict = {
        key: value
        for key, value in color.items()
        if isinstance(value, tuple) and 3 <= len(value) <= 4 and all(0 <= x <= 255 for x in value)
    }
    return processed_dict or {}



import asyncio
from PIL import Image, ImageDraw
from typing import Tuple
from ..model.base import Translator
from ..model.api import Character, Equipment, PlayerData
from ..tools.git import ImageCache
from ..tools import pill
from ..model.generator import ZZZCard
from ..model.base import URL_ASSETS
from ..model.assets import get_agent_icon, get_color_agent

_of = ImageCache()

async def get_skill_icon(index: int) -> Image.Image:
    return {
        0: await _of.skill_1_s,
        1: await _of.skill_4_s,
        2: await _of.skill_2_s,
        3: await _of.skill_5_s,
        5: await _of.skill_6_s,
        6: await _of.skill_3_s,
    }.get(index, await _of.skill_1_s)


async def get_stars(index: int) -> Image.Image:
    return {
        1: await _of.g_one,
        2: await _of.g_two,
        3: await _of.g_three,
        4: await _of.g_four,
        5: await _of.g_five,
    }.get(index, await _of.g_one)


async def get_rarity(index: int) -> Image.Image:
    return {
        2: await _of.rank_w_b,
        3: await _of.rank_w_a,
        4: await _of.rank_w_s,
    }.get(index, await _of.rank_w_b)

async def get_rank_color(index: int) -> Tuple[int,int,int,int]:
    return {
        2: (7,158,254),
        3: (181,7,254),
        4: (254,181,7)
    }.get(index, (254,181,7))

async def get_core(index: int) -> Image.Image:
    return {
        0: await _of.core_skill_a,
        1: await _of.core_skill_b,
        2: await _of.core_skill_c,
        3: await _of.core_skill_d,
        4: await _of.core_skill_e,
        5: await _of.core_skill_f,
    }.get(index, await _of.core_skill_a)


async def get_cinema(index: int) -> Image.Image:
    return {
        0: await _of.conts_1_lvl,
        1: await _of.conts_2_lvl,
        2: await _of.conts_3_lvl,
        3: await _of.conts_4_lvl,
        4: await _of.conts_5_lvl,
        5: await _of.conts_6_lvl,
    }.get(index, await _of.conts_1_lvl)


class StyleTwo:
    def __init__(self, data: Character, player: PlayerData, lang: Translator, art: str = None, color: tuple = None, hide: bool = False, crop: int = 0):
        self.data = data
        self.player = player
        self.lang = lang
        self.art = art
        self.hide = hide
        self.crop = crop
        
        if color:
            self.color = color
        elif art:
            self.color = None
        else:
            colors = get_color_agent(data.id)
            self.color = colors.mindscape

        _of.set_mapping(3)


    async def create_background(self):
        background = Image.new("RGBA", (457,809), (0,0,0,0))
        image_background = Image.new("RGBA", (457,809), (0,0,0,0))
        text_frame = await _of.text_frame
        background = await _of.background
        mask = await _of.background_mask
        self.charter_bg = await _of.background
        self.charter_bg = self.charter_bg.convert("RGBA").copy()
        if self.art:
            self.art = await pill.get_user_image(self.art)
            user_image = await pill.get_center_size((502,809), self.art)
            if not self.color:
                self.color = await pill.get_colors(user_image.convert("RGBA"), 5, common=True, radius=30, quality=800)
                ll = await pill.light_level(self.color[:3])
                if ll < 45:
                    self.color = await pill.get_light_pixel_color(self.color[:3],up = True)
            image_background.alpha_composite(user_image,(-22,0))

            background.paste(image_background,(0,0), mask.convert("L"))
            self.charter_bg.alpha_composite(background)
        else:
            text_frame = await pill.recolor_image(text_frame, self.color[:3])
            self.charter_bg.alpha_composite(text_frame)
            icon = get_agent_icon(self.data.id)
            user_image = await pill.get_download_img(icon.image, size=(843,879))
            user_image_w = await pill.recolor_image(user_image, (255,255,255))
            user_image_b = await pill.recolor_image(user_image, (0,0,0))

            image_background.alpha_composite(user_image_w, (-186,0))
            image_background.alpha_composite(user_image_b, (-198,0))
            image_background.alpha_composite(user_image, (-193, 0))
            background.paste(image_background,(0,0), mask.convert("L"))
            self.charter_bg.alpha_composite(background)
        
    async def create_name(self):
        self.name = Image.new("RGBA", (491,142), (0,0,0,0))

        name_main = await pill.create_image_with_text(self.data.name, 69, stroke_width=2, stroke_fill= (0,0,0,255), max_width= 492)
        name = await pill.create_image_with_text(self.data.name, 69, max_width= 492)
        name = await pill.recolor_image(name, self.color[:3]) 
        lvl = f"{self.lang.lvl}: {self.data.level}/{self.data.max_level*10}"
        lvl_main = await pill.create_image_with_text(lvl, 50, stroke_width=2, stroke_fill= (0,0,0,255) , max_width= 492)
        lvl = await pill.create_image_with_text(lvl, 50, max_width= 492)
        lvl = await pill.recolor_image(lvl, self.color[:3]) 
        self.name.alpha_composite(name, (13,9))
        self.name.alpha_composite(name_main, (11,5))
        self.name.alpha_composite(lvl, (17,87))
        self.name.alpha_composite(lvl_main, (15,83))
        self.name = self.name.resize((191,57))
    
    async def create_skill(self):
        skill = await _of.skill_bg
        self.skill = skill.copy()
        skill_coont = await _of.skill_lvl
        skill_frame = await _of.skill_frame
        skill_frame = await pill.recolor_image(skill_frame, self.color[:3])
        font = await pill.get_font(18)
        last_true_index = next((i for i in reversed(range(len(self.data.cinema))) if self.data.cinema[i]), 0)

        position_skill = [
            (8,0),
            (79,0),
            (150,0),
            (221,0),
            (292,0),
            (363,0),
        ]

        for i, key in enumerate(self.data.skills):
            skill_icon = Image.new("RGBA", (63,70), (0,0,0,0))
            count = skill_coont.copy()
            icon = await get_skill_icon(key.index)
            lvl = key.level
            if last_true_index >= 3 and i != 4:
                lvl += 2
            if last_true_index >= 5 and i != 4:
                lvl += 2
                
            d = ImageDraw.Draw(count)
            x = int(font.getlength(str(lvl)) / 2)
            d.text((13 - x, 0), str(lvl), font=font, fill=self.color)
            
            skill_icon.alpha_composite(icon.resize((59,59)), (2,8))
            skill_icon.alpha_composite(skill_frame)
            skill_icon.alpha_composite(count,(19,0))
            
            self.skill.alpha_composite(skill_icon, position_skill[i])

    async def create_info(self):
        info = await _of.info_bg
        self.info = info.copy()

        icon_literal = None
        for i in range(6):
            icon = await get_core(i)
            icon = icon.resize((51,51))
            if self.data.core_skill-1 >= i:
                icon_literal = await pill.recolor_image(icon, self.color[:3])
        
        if not icon_literal:
            icon = await get_core(0)
            icon_literal = await pill.recolor_image(icon, (103,103,103))

        self.info.alpha_composite(icon_literal.resize((42,42)), (3,97))
        
        icon_const = await pill.recolor_image(await _of.conts_0_lvl, self.color[:3])
        for i in range(6):
            icon = await get_cinema(i)
            if i < self.data.const and self.data.const != 0:
                icon_const = await pill.recolor_image(icon, self.color[:3])

        self.info.alpha_composite(icon_const.resize((42,42)), (3,147))

        element_icon = await pill.get_download_img(self.data.element.icon, size = (42,42))
        self.info.alpha_composite(element_icon, (3,2))
        fraction_icon = await pill.get_download_img(self.data.profession.icon, size = (42,42))
        self.info.alpha_composite(fraction_icon, (3,48))

    async def create_weapon(self):
        weapon = await _of.weapon_bg
        self.weapon = weapon.copy()
        if not self.data.weapon:
            return

        icon = await pill.get_download_img(URL_ASSETS + self.data.weapon.icon, (143,143))
        icon_color = await pill.recolor_image(icon, self.color[:3]) 
        icon_w = await pill.recolor_image(icon, (255,255,255))

        self.weapon.alpha_composite(icon_w, (168,9))
        self.weapon.alpha_composite(icon_color, (176,0))
        self.weapon.alpha_composite(icon, (172,3))

        rarity = await get_rarity(self.data.weapon.rarity)
        self.weapon.alpha_composite(rarity.convert("RGBA").resize((35,35)), (269,100))

        lvl_bg = await _of.levl
        lvl_bg = lvl_bg.copy()

        font = await pill.get_font(24)
        d = ImageDraw.Draw(lvl_bg)
        text_lvl = f"LVL: {self.data.weapon.level}"
        x = int(font.getlength(text_lvl) / 2)
        d.text((47 - x, 3), text_lvl, font=font, fill=self.color)
        x = int(font.getlength(text_lvl) / 2)
        d.text((43 - x, 0), text_lvl, font=font, fill=(255,255,255,255), stroke_width=2, stroke_fill=(0,0,0,255))
        self.weapon.alpha_composite(lvl_bg, (315,76))

        font = await pill.get_font(22)
        up_bg = await _of.up
        up_bg = up_bg.copy()
        d = ImageDraw.Draw(up_bg)
        text_up = f"R{self.data.weapon.cons}"
        x = int(font.getlength(text_up) / 2)
        d.text((17 - x, 3), text_up, font=font, fill=(37,37,37,255))
        self.weapon.alpha_composite(up_bg, (411,76))


        icon_main = await pill.get_download_img(self.data.weapon.main.icon, size=(30,30))
        icon_sub = await pill.get_download_img(self.data.weapon.sub.icon, size=(30,30))
        font = await pill.get_font(23)
        d = ImageDraw.Draw(self.weapon)

        self.weapon.alpha_composite(icon_main, (315,41))
        self.weapon.alpha_composite(icon_sub, (407,41))

        d.text((349, 39), str(self.data.weapon.main.value), font=font, fill=(255,255,255,255))
        formatted_percentage = self.data.weapon.sub.get_value()
        d.text((442, 39), str(formatted_percentage), font=font, fill=(255,255,255,255))


        name_main = await pill.create_image_with_text(self.data.weapon.name, 25, stroke_width=2, stroke_fill= (0,0,0,255), max_width= 159)
        name = await pill.create_image_with_text(self.data.weapon.name, 25, max_width= 159) 
        name = await pill.recolor_image(name, self.color[:3]) 
        xyz = 63
        x = 14
        if name_main.size[1] > 60:
            xyz = 42
            x = 40
        self.weapon.alpha_composite(name, (x,xyz))
        self.weapon.alpha_composite(name_main, (x-2,xyz-2))

    async def create_stats(self):
        stats = await _of.stats_bg
        self.stats = stats.copy()
        position_text_x = 158
        position_text_y = 13

        position_icon_x = 68
        position_icon_y = 12

        line = Image.new("RGBA", (2,204), self.color)
        self.stats.alpha_composite(line, (164,13))

        font = await pill.get_font(20)
        d = ImageDraw.Draw(self.stats)
        
        for i, key in enumerate(self.data.stats):
            icon = await pill.get_download_img(key.icon, (25,25))
            self.stats.alpha_composite(icon, (position_icon_x,position_icon_y))
            text_val = key.get_value()
            x = int(font.getlength(text_val))
            d.text((position_text_x - x, position_text_y), text_val, font=font, fill= (255,255,255,255))

            position_icon_y += 36
            position_text_y += 36

            if i == 5:
                position_text_x = 261
                position_text_y = 13

                position_icon_x = 171
                position_icon_y = 12

    async def create_disc(self, data: Equipment):
        echo_background = Image.new("RGBA", (710,85), (0,0,0,0))
        
        background_disk = await _of.disk_bg
        background_disk = background_disk.copy()
        icon_disc = await pill.get_download_img(URL_ASSETS + data.icon, (74,74))
        color = await pill.get_colors(icon_disc.convert("RGBA"), 15, common=True, radius=5, quality=800)

        background_disk.alpha_composite(icon_disc, (5,5))
        rarity = await get_rarity(data.rarity)
        rank_leve = await _of.rank_leve
        color_rank = await get_rank_color(data.rarity)
        rank_leve = await pill.recolor_image(rank_leve,color_rank)
        background_disk.alpha_composite(rank_leve.resize((93,100)), (1,1))
        background_disk.alpha_composite(rarity.copy().convert("RGBA").resize((22,22)), (0,5))

        lvl_bg = await _of.lvl_disk
        lvl_bg = lvl_bg.copy()
        lvl_frame = await _of.frame_disk_lvl
        lvl_frame = await pill.recolor_image(lvl_frame,color_rank)
        font = await pill.get_font(16)
        d = ImageDraw.Draw(lvl_bg)
        x = int(font.getlength(f"+{data.level}") / 2)
        d.text((20-x, 0), f"+{data.level}", font=font, fill=self.color)
        lvl_bg.alpha_composite(lvl_frame)
        background_disk.alpha_composite(lvl_bg, (14,61))

        echo_background.alpha_composite(background_disk, (0,0))


        stats_frame = Image.new("RGBA", (224,71),(0,0,0,0))

        main_bg_disk_stat = await _of.main_bg_disk_stat
        main_bg_disk_stat = main_bg_disk_stat.copy()
        main_bg_disk_stat = await pill.recolor_image(main_bg_disk_stat, color[:3])
        texture_disk_main = await _of.texture_disk_main
        main_bg_disk_stat.alpha_composite(texture_disk_main)

        sub_bg_disk_stat = await _of.sub_bg_disk_stat
        sub_bg_disk_stat = sub_bg_disk_stat.copy()

        stats_frame.alpha_composite(sub_bg_disk_stat)
        stats_frame.alpha_composite(main_bg_disk_stat)

        d = ImageDraw.Draw(stats_frame)
        icon_main = await pill.get_download_img(data.main[0].icon, (25,25))
        color_main = (255,255,255,255) if pill.is_white_visible(color[:3]) else (26,26,26,255)
        icon_main = await pill.recolor_image(icon_main, color_main[:3])

        stats_frame.alpha_composite(icon_main,(54,11))
        value = str(data.main[0].get_value())
        font = await pill.get_font(21)
        x = int(font.getlength(value))
        d.text((79-x, 43), value, font=font, fill=color_main)


        position = [
            (89,4),
            (103,37),
            (154,4),
            (167,37),
        ]
        ll = await pill.light_level(color[:3])
        if ll < 45:
            color = await pill.get_light_pixel_color(color[:3],up = True)

        rolls_line =  Image.new("RGBA", (7,2), color)
        font = await pill.get_font(14)
        for i, key in enumerate(data.sub):
            sub_bg = Image.new("RGBA", (55,21), (0,0,0,0))
            sub_icon = await pill.get_download_img(key.icon, (17,17))
            sub_bg.alpha_composite(sub_icon)
            value = str(key.get_value())
            d = ImageDraw.Draw(sub_bg)
            d.text((21, 0), value, font=font, fill=(255,255,255,255))
            x = 2
            for _ in range(key.rolls-1):
                sub_bg.alpha_composite(rolls_line, (x, 19))
                x += 15
            stats_frame.alpha_composite(sub_bg, position[i])

        echo_background.alpha_composite(stats_frame, (486,7))

        return echo_background


    async def build(self):
        self.background = Image.new("RGBA", (710,923), (0,0,0,0))

        y = 57
        for key in self.disc:
            self.background.alpha_composite(key, (0,y))
            y += 86
            
        self.background.alpha_composite(self.skill,(61,0))
        self.background.alpha_composite(self.info,(0,578))
        self.background.alpha_composite(self.stats,(443,573))
        self.background.alpha_composite(self.charter_bg,(47,44))
        self.background.alpha_composite(self.name, (56,741))
        self.background.alpha_composite(self.weapon, (23,772))

    
    async def start(self) -> ZZZCard:
        await self.create_background()

        task = [
            self.create_name(),
            self.create_skill(),
            self.create_info(),
            self.create_weapon(),
            self.create_stats(),
        ]

        await asyncio.gather(*task)

        self.disc = []
        self.sets = {}
        for key in self.data.equippe:
            self.disc.append(await self.create_disc(key.equipment))
            if not key.equipment.sets.id in self.sets:
                self.sets[key.equipment.sets.id] = {"count": 1, "name": key.equipment.sets.name, "icon": key.equipment.sets.icon}
            else:
                self.sets[key.equipment.sets.id]["count"] += 1

        await self.build()

        return ZZZCard(
            id = self.data.id,
            name = self.data.name,
            color = self.color,
            icon = self.data.icon.circle_icon,
            card = self.background,
        )
        


# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.
import json
from PIL import ImageFont,Image,ImageDraw
from .. import cache, git

_caches = cache.Cache.get_cache()

async def get_font(size):
    return ImageFont.truetype(git.font, size)

async def get_text_size_frame(text,font_size,frame_width):    
    font = await get_font(font_size)

    while font.getlength(text) > frame_width:
        font_size -= 1
        font = await get_font(font_size)
    return font,font.getlength(text)


async def create_image_with_text(text: str, font_size: int, max_width: int =336, color: tuple =(255, 255, 255, 255), stroke_width: int = 0, stroke_fill: tuple = None, alg="Left"):
    cache_key = json.dumps((text, font_size, max_width, color, alg), sort_keys=True)
    if cache_key in _caches:
        return _caches[cache_key]
    
    font = await get_font(font_size)

    lines = []
    line = []
    for word in text.split():
        if line:
            temp_line = line + [word]
            temp_text = ' '.join(temp_line)
            temp_width = font.getmask(temp_text).getbbox()[2]
            if temp_width <= max_width:
                line = temp_line
            else:
                lines.append(line)
                line = [word]
        else:
            line = [word]
    if line:
        lines.append(line)

    width = 0
    height = 0
    for line in lines:
        line_width = font.getmask(' '.join(line)).getbbox()[2]
        width = max(width, line_width)
        height += font.getmask(' '.join(line)).getbbox()[3]

    img = Image.new('RGBA', (min(width, max_width) + 10, height + (font_size)), color=(255, 255, 255, 0))

    draw = ImageDraw.Draw(img)
    
    y_text = 0
    for line_num, line in enumerate(lines):
        text_width, text_height = font.getmask(' '.join(line)).getbbox()[2:]
        if alg == "center" and line_num > 0:
            x_text = (max_width - text_width) // 2
        else:
            x_text = 0
        draw.text((x_text, y_text), ' '.join(line), font=font, fill=color, stroke_width = stroke_width, stroke_fill= stroke_fill)
        y_text += text_height + 5
        
    _caches[cache_key] = img
    
    return img


from pydantic import BaseModel
from typing import Tuple
from ..tools.api import DataManager
from ..model.base import URL_ASSETS

ICON = "https://act-webstatic.hoyoverse.com/game_record/zzzv2/role_square_avatar/role_square_avatar_{agent_id}.png"

def hex_to_rgba(hex_code: str) -> Tuple[int,int,int,int]:
    hex_code = hex_code.strip('#')
    
    red = int(hex_code[0:2], 16)
    green = int(hex_code[2:4], 16)
    blue = int(hex_code[4:6], 16)
    
    alpha = 255 
    
    return red, green, blue, alpha

class IconAssets(BaseModel):
    icon: str = None
    image: str = None
    full: str = None
    circle_icon: str = None
    
class ColorAssets(BaseModel):
    accent: Tuple[int,int,int,int]
    mindscape: Tuple[int,int,int,int]

def get_agent_icon(id: int) -> IconAssets:
    data = DataManager().get_data()
    return IconAssets(
        icon = ICON.format(agent_id = id),
        image = data.get("hoyolink_data").get(str(id)).get("link"),
        full = URL_ASSETS + data.get("avatars_data").get(str(id)).get("Image"),
        circle_icon = URL_ASSETS + data.get("avatars_data").get(str(id)).get("CircleIcon"),
    )


def get_color_agent(id: int) -> ColorAssets:
    data = DataManager().get_data()
    return ColorAssets(
        accent= hex_to_rgba(data.get("avatars_data").get(str(id)).get("Colors").get("Accent")),
        mindscape= hex_to_rgba(data.get("avatars_data").get(str(id)).get("Colors").get("Mindscape"))
    )